#!/usr/bin/env python3
###########################
### query_processing.py ###
###########################
# Author: Samuel Aroney

import polars as pl
import os
import logging
import argparse

QUERY_COLUMNS = {
    "query_name": str,
    "query_sequence": str,
    "divergence": int,
    "num_hits": int,
    "coverage": float,
    "sample": str,
    "marker": str,
    "hit_sequence": str,
    "taxonomy": str,
}

PIPE_COLUMNS = {
    "gene": str,
    "sample": str,
    "sequence": str,
    "num_hits": int,
    "coverage": int,
    "taxonomy": str,
}

OUTPUT_COLUMNS={
    "gene": str,
    "sample": str,
    "sequence": str,
    "num_hits": int,
    "coverage": int,
    "taxonomy": str,
    "found_in": str,
}

# Example shell directive for Snakemake:
# shell:
# """
# python3 binchicken/workflow/scripts/query_processing.py \
#   --query-reads {input.query_reads} \
#   --pipe-reads {input.pipe_reads} \
#   --binned {output.binned} \
#   --unbinned {output.unbinned} \
#   --sequence-identity {params.sequence_identity} \
#   --window-size {params.window_size} \
#   --threads {threads} \
#   --log {log}
# """

def processing(
    query_read,
    pipe_read,
    SEQUENCE_IDENTITY=0.86,
    WINDOW_SIZE=60):

    if len(query_read) == 0:
        empty_output = pl.DataFrame(schema=OUTPUT_COLUMNS)
        return empty_output, empty_output

    appraised = (
        query_read
        # Rename to match appraise output
        # Query output: query_name, query_sequence, divergence, num_hits, coverage, sample, marker, hit_sequence, taxonomy
        # Appraise output: gene, sample, sequence, num_hits, coverage, taxonomy, found_in
        .rename({"marker": "gene", "query_name":"sample", "query_sequence": "sequence", "sample": "found_in"})
        # Query taxonomy, num_hits and coverage are from database (i.e. the reference genomes)
        .drop(["taxonomy", "num_hits", "coverage"])
        .join(pipe_read, on=["gene", "sample", "sequence"], how="inner")
        .group_by(["gene", "sample", "sequence", "num_hits", "coverage", "taxonomy", "divergence"])
        .agg(pl.col("found_in").sort().str.concat(","))
        .with_columns(pl.col("divergence").alias("binned") <= ((1 - SEQUENCE_IDENTITY) * WINDOW_SIZE))
    )

    # Split dataframe into binned/unbinned
    binned = appraised.filter(pl.col("binned")).drop(["divergence", "binned"])
    unbinned = (
        pipe_read
        .join(
            appraised, on=["gene", "sample", "sequence", "num_hits", "coverage", "taxonomy"], how="left", coalesce=True
            )
        .filter(~pl.col("binned").fill_null(False))
        .drop(["divergence", "binned"])
        .with_columns(pl.lit(None).cast(str).alias("found_in"))
    )

    return binned, unbinned

def pipeline(
    query_reads,
    pipe_reads,
    SEQUENCE_IDENTITY=0.86,
    WINDOW_SIZE=60):

    logging.info(f"Polars using {str(pl.thread_pool_size())} threads")

    for query, pipe in zip(query_reads, pipe_reads):
        logging.debug(f"Processing {query} and {pipe}")
        binned, unbinned = processing(
            pl.read_csv(query, separator="\t"),
            pl.read_csv(pipe, separator="\t"),
            SEQUENCE_IDENTITY,
            WINDOW_SIZE)
        yield binned, unbinned

def main():
    parser = argparse.ArgumentParser(description="Query processing pipeline script.")
    parser.add_argument("--query-reads", required=True, nargs='+', help="List of query read files")
    parser.add_argument("--pipe-reads", required=True, nargs='+', help="List of pipe read files")
    parser.add_argument("--binned", required=True, help="Path to output binned file")
    parser.add_argument("--unbinned", required=True, help="Path to output unbinned file")
    parser.add_argument("--sequence-identity", type=float, default=0.86, help="Sequence identity threshold")
    parser.add_argument("--window-size", type=int, default=60, help="Window size")
    parser.add_argument("--threads", type=int, default=1, help="Number of threads for Polars")
    parser.add_argument("--log", default=None, help="Log file path")
    args = parser.parse_args()

    os.environ["POLARS_MAX_THREADS"] = str(args.threads)
    import polars as pl

    if args.log:
        logging.basicConfig(
            filename=args.log,
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )
    else:
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )

    outputs = pipeline(
        args.query_reads,
        args.pipe_reads,
        SEQUENCE_IDENTITY=args.sequence_identity,
        WINDOW_SIZE=args.window_size
    )

    with open(args.binned, "ab") as binned_file, open(args.unbinned, "ab") as unbinned_file:
        first = True
        for binned, unbinned in outputs:
            binned.write_csv(binned_file, separator="\t", include_header=first)
            unbinned.write_csv(unbinned_file, separator="\t", include_header=first)
            first = False

    logging.info("Done")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
########################
### cluster_graph.py ###
########################
# Author: Samuel Aroney

import polars as pl
import itertools
import os
import logging
import argparse

OUTPUT_COLUMNS={
    "samples": str,
    "length": int,
    "total_targets": float,
    "total_size": int,
    "recover_samples": str,
    "coassembly": str,
    }

TARGET_WEIGHTING_COLUMNS = {
    "target": str,
    "weight": float,
}

# Example shell directive for Snakemake:
# shell:
# """
# python3 binchicken/workflow/scripts/cluster_graph.py \
#   --elusive-edges {input.elusive_edges} \
#   --read-size {input.read_size} \
#   --targets-weighted {input.targets_weighted} \
#   --elusive-clusters {output.elusive_clusters} \
#   --max-coassembly-size {params.max_coassembly_size} \
#   --max-coassembly-samples {params.max_coassembly_samples} \
#   --num-coassembly-samples {params.num_coassembly_samples} \
#   --max-recovery-samples {params.max_recovery_samples} \
#   --max-samples-combinations {params.max_samples_combinations} \
#   --coassembly-samples {input.coassembly_samples} \
#   --exclude-coassemblies {input.exclude_coassemblies} \
#   --single-assembly {params.single_assembly} \
#   --anchor-samples {input.anchor_samples} \
#   --threads {threads} \
#   --log {log}
# """

def join_list_subsets(df1, df2):
    """
    Join two DataFrames/LazyFrames by strict subset of list column
    """
    df2 = (
        df2.select(
            right_samples = pl.col("samples"),
            extra_targets = pl.col("target_ids"),
            length = pl.col("cluster_size").cast(pl.UInt32),
            )
    )

    output = (
        df1
        .select("samples", "samples_hash", "length")
        .join(df2, on="length")
        .filter(pl.col("right_samples").is_not_null())
        .filter(pl.col("samples").list.set_intersection("right_samples").list.len() >= pl.col("length"))
        .group_by("samples_hash")
        .agg(pl.col("extra_targets").flatten())
        .select(
            pl.col("samples_hash"),
            pl.col("extra_targets").list.unique().fill_null([]),
            )
    )

    return (
        df1
        .join(output, on="samples_hash", how="left", coalesce=True)
        .with_columns(pl.col("extra_targets").fill_null([]))
        )

def accumulate_clusters(x):
    clustered_samples = {}
    choices = []
    for cluster in x:
        cluster_size = len(cluster)
        if cluster_size not in clustered_samples:
            clustered_samples[cluster_size] = set()
        if all([s not in clustered_samples[cluster_size] for s in cluster]):
            choices.append(True)
            clustered_samples[cluster_size].update(cluster)
        else:
            choices.append(False)

    return pl.Series(choices, dtype=pl.Boolean)

def find_recover_candidates(df, samples_df, MAX_RECOVERY_SAMPLES=20):
    samples_df = samples_df.explode("target_ids")

    output = (
        df
        .select("samples_hash", "target_ids")
        .explode("target_ids")
        .join(samples_df, on="target_ids", how="left", coalesce=True)
        .group_by("samples_hash", "recover_candidates")
        .agg(pl.len().alias("count"))
        .sort("count", descending=True)
        .group_by("samples_hash")
        .head(MAX_RECOVERY_SAMPLES)
        .group_by("samples_hash")
        .agg("recover_candidates")
    )

    return df.join(output, on="samples_hash", how="left", coalesce=True)

def pipeline(
        elusive_edges,
        read_size,
        weightings=None,
        anchor_samples=set(),
        MAX_COASSEMBLY_SIZE=None,
        MAX_COASSEMBLY_SAMPLES=2,
        MIN_COASSEMBLY_SAMPLES=2,
        MAX_RECOVERY_SAMPLES=20,
        MIN_CLUSTER_TARGETS=1,
        MAX_SAMPLES_COMBINATIONS=100,
        COASSEMBLY_SAMPLES=[],
        EXCLUDE_COASSEMBLIES=[],
        single_assembly=False):

    logging.info(f"Polars using {str(pl.thread_pool_size())} threads")

    if len(elusive_edges) == 0:
        logging.warning("No elusive edges found")
        return pl.DataFrame(orient="row", schema=OUTPUT_COLUMNS)

    is_pooled = any(elusive_edges["style"] == "pool")

    with pl.StringCache():
        elusive_edges = (
            elusive_edges
            .with_columns(
                pl.col("samples")
                    .str.split(",")
                    .cast(pl.List(pl.Categorical)),
                pl.col("target_ids")
                    .str.split(",")
                    .cast(pl.List(pl.UInt64)),
                )
            .with_columns(
                samples_hash = pl.col("samples").list.sort().hash(),
                length = pl.col("samples").list.len()
                )
        )

        if weightings is not None:
            if weightings.height == 0:
                logging.error("No target weightings found")
                return pl.DataFrame(schema=OUTPUT_COLUMNS)

            weightings = (
                weightings
                .select(target_ids = pl.col("target").cast(pl.UInt64), weight = "weight")
            )
            weightings_dict = dict(weightings.iter_rows())
        else:
            weightings_dict = {}

        if COASSEMBLY_SAMPLES:
            coassembly_edges = (
                elusive_edges
                .with_columns(
                    pl.col("samples").list.eval(pl.element().filter(pl.element().is_in(COASSEMBLY_SAMPLES)))
                    )
                .with_columns(length = pl.col("samples").list.len())
                .filter(pl.col("samples").list.len() >= pl.col("cluster_size"))
            )
        else:
            coassembly_edges = elusive_edges

        read_size = (
            read_size
            .select(
                "read_size",
                samples = pl.col("sample").cast(pl.Categorical),
                )
        )

        if EXCLUDE_COASSEMBLIES:
            excluded_coassemblies = (
                pl.DataFrame({"samples": EXCLUDE_COASSEMBLIES})
                .with_columns(
                    samples_hash = pl.col("samples")
                        .str.split(",")
                        .cast(pl.List(pl.Categorical))
                        .list.sort().hash()
                    )
            )
        else:
            excluded_coassemblies = pl.DataFrame(schema={"samples_hash": pl.UInt64})

        if MAX_COASSEMBLY_SAMPLES == 1:
            logging.info("Skipping clustering, using single-sample clusters")
            clusters = [
                elusive_edges
                .explode("samples")
                .filter((not COASSEMBLY_SAMPLES) | pl.col("samples").is_in(COASSEMBLY_SAMPLES))
                .group_by("samples")
                .agg(pl.col("target_ids").flatten())
                .with_columns(
                    pl.concat_list(pl.col("samples")),
                    pl.col("target_ids").list.sort().list.unique(),
                    samples_hash = pl.concat_list(pl.col("samples")).list.sort().hash(),
                )
            ]
        else:
            logging.info("Forming candidate sample clusters")
            clusters = [
                coassembly_edges
                .filter(pl.col("style") == "match")
                .filter(pl.col("cluster_size") >= MIN_COASSEMBLY_SAMPLES)
                .filter(pl.col("target_ids").list.len() >= MIN_CLUSTER_TARGETS)
                .select("samples", "target_ids", samples_hash = pl.col("samples").list.sort().hash())
            ]

            if is_pooled:
                clusters.append(
                    coassembly_edges
                    .filter(pl.col("style") == "pool")
                    .filter(pl.col("cluster_size") >= MIN_COASSEMBLY_SAMPLES)
                    # Prevent combinatorial explosion (also, large clusters are less useful for distinguishing between clusters)
                    .filter(pl.col("samples").list.len() <= MAX_SAMPLES_COMBINATIONS)
                    .with_columns(
                        samples_combinations = pl.struct(["length", "cluster_size"]).map_elements(
                            lambda x: [i for i in itertools.combinations(range(x["length"]), x["cluster_size"])],
                            return_dtype=pl.List(pl.List(pl.Int64)),
                            )
                        )
                    .explode("samples_combinations")
                    .with_columns(pl.col("samples").list.gather(pl.col("samples_combinations")))
                    .with_columns(samples_hash = pl.col("samples").list.sort().hash())
                    .select("samples_hash", "samples", "target_ids", "cluster_size")
                    .group_by("samples_hash")
                    .agg(
                        pl.first("samples"),
                        pl.col("target_ids").flatten(),
                        pl.first("cluster_size"),
                        )
                    .filter(pl.col("target_ids").list.len() >= MIN_CLUSTER_TARGETS)
                    .select("samples", "target_ids", "samples_hash")
                )

        sample_targets = (
            elusive_edges
            .select("target_ids", recover_candidates = pl.col("samples"))
            .explode("recover_candidates")
            .explode("target_ids")
            .unique()
            .group_by("recover_candidates")
            .agg("target_ids")
        )

        def filter_max_coassembly_size(df, MAX_COASSEMBLY_SIZE):
            if MAX_COASSEMBLY_SIZE is None:
                return df
            else:
                return df.filter(pl.col("total_size") <= MAX_COASSEMBLY_SIZE)

        logging.info("Filtering clusters (each sample restricted to only once per cluster size)")
        clusters = (
            pl.concat(clusters)
            .join(excluded_coassemblies, on="samples_hash", how="anti")
            .filter((not anchor_samples) | (pl.col("samples").list.eval(pl.element().is_in(anchor_samples)).list.any()))
            .with_columns(length = pl.col("samples").list.len())
            .explode("samples")
            .join(read_size, on="samples", how="left", coalesce=True)
            .group_by("samples_hash")
            .agg(
                pl.col("samples"),
                pl.first("length"),
                pl.first("target_ids"),
                total_size = pl.sum("read_size"),
                )
            .pipe(
                filter_max_coassembly_size,
                MAX_COASSEMBLY_SIZE=MAX_COASSEMBLY_SIZE,
                )
            .with_columns(weighting = pl.lit(weightings is not None))
            .with_columns(
                total_targets = pl.when(pl.col("weighting"))
                .then(pl.col("target_ids").list.eval(pl.element().replace_strict(weightings_dict, default=0)).list.sum())
                .otherwise(pl.col("target_ids").list.len()),
            )
            .sort("total_targets", "total_size", descending=[True, False])
            .with_columns(
                unique_samples = 
                    pl.col("samples")
                    .map_batches(accumulate_clusters, return_dtype=pl.Boolean),
                )
            .filter(pl.col("unique_samples"))
            .drop("unique_samples")
            .pipe(
                join_list_subsets,
                df2=coassembly_edges
                    .filter(pl.col("style") == "pool")
                    .filter(pl.col("samples").list.len() > MAX_SAMPLES_COMBINATIONS),
                )
            .with_columns(pl.concat_list("target_ids", "extra_targets").list.unique())
            .pipe(
                find_recover_candidates,
                sample_targets,
                MAX_RECOVERY_SAMPLES=MAX_RECOVERY_SAMPLES,
                )
            .with_columns(
                pl.col("samples").cast(pl.List(pl.Utf8)).list.sort().list.join(","),
                recover_samples = pl.concat_list(pl.col("samples"), "recover_candidates")
                .list.unique(maintain_order=True)
                .list.head(MAX_RECOVERY_SAMPLES)
                .cast(pl.List(pl.Utf8))
                .list.sort()
                .list.join(","),
                total_targets = pl.when(pl.col("weighting"))
                    .then(pl.col("total_targets"))
                    .otherwise(pl.col("target_ids").list.len()),
                )
            .sort("total_targets", "total_size", descending=[True, False])
            .with_row_index("coassembly")
            .select(
                "samples", "length", "total_targets", "total_size", "recover_samples",
                coassembly = pl.when(single_assembly)
                    .then(pl.col("samples"))
                    .otherwise(pl.lit("coassembly_") + pl.col("coassembly").cast(pl.Utf8))
                )
        )

    logging.info("Done")

    return clusters

def main():
    parser = argparse.ArgumentParser(description="Cluster graph pipeline script.")
    parser.add_argument("--elusive-edges", required=True, help="Path to elusive edges input file")
    parser.add_argument("--read-size", required=True, help="Path to read size input file")
    parser.add_argument("--targets-weighted", nargs='?', const=None, default=None, help="Path to targets weighted input file (optional)")
    parser.add_argument("--elusive-clusters", required=True, help="Path to output elusive clusters file")
    parser.add_argument("--max-coassembly-size", type=float, nargs='?', const=None, default=None, help="Max coassembly size in GB (optional)")
    parser.add_argument("--max-coassembly-samples", type=int, default=2, help="Max coassembly samples")
    parser.add_argument("--num-coassembly-samples", type=int, default=2, help="Num coassembly samples (min)")
    parser.add_argument("--max-recovery-samples", type=int, default=20, help="Max recovery samples")
    parser.add_argument("--max-samples-combinations", type=int, default=100, help="Max samples combinations for pooled clusters")
    parser.add_argument("--coassembly-samples", nargs='?', default=None, help="List file of coassembly samples")
    parser.add_argument("--exclude-coassemblies", nargs='?', default=None, help="List file of excluded coassemblies")
    parser.add_argument("--single-assembly", type=lambda x: x.lower() == 'true', nargs='?', const=True, default=False, help="Single assembly mode (True/False)")
    parser.add_argument("--anchor-samples", nargs='?', default=None, help="List file of anchor samples")
    parser.add_argument("--threads", type=int, default=1, help="Number of threads for Polars")
    parser.add_argument("--log", default=None, help="Log file path")
    args = parser.parse_args()

    os.environ["POLARS_MAX_THREADS"] = str(args.threads)
    import polars as pl

    if args.log:
        logging.basicConfig(
            filename=args.log,
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )
    else:
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )

    MAX_COASSEMBLY_SIZE = args.max_coassembly_size * 10**9 if args.max_coassembly_size else None
    MAX_COASSEMBLY_SAMPLES = args.max_coassembly_samples
    MIN_COASSEMBLY_SAMPLES = args.num_coassembly_samples
    MAX_RECOVERY_SAMPLES = args.max_recovery_samples
    MAX_SAMPLES_COMBINATIONS = args.max_samples_combinations
    COASSEMBLY_SAMPLES = pl.read_csv(args.coassembly_samples, has_header=False, new_columns=["sample"]).get_column("sample").to_list() if args.coassembly_samples else []
    EXCLUDE_COASSEMBLIES = pl.read_csv(args.exclude_coassemblies, separator="\t", has_header=False, new_columns=["coassembly"]).get_column("coassembly").to_list() if args.exclude_coassemblies else []
    single_assembly = args.single_assembly
    anchor_samples = set(pl.read_csv(args.anchor_samples, has_header=False, new_columns=["sample"]).get_column("sample").to_list()) if args.anchor_samples else set()

    elusive_edges = pl.read_csv(args.elusive_edges, separator="\t", schema_overrides={"target_ids": str})
    read_size = pl.read_csv(args.read_size, has_header=False, new_columns=["sample", "read_size"])

    if args.targets_weighted:
        weightings = pl.read_csv(args.targets_weighted, separator="\t", schema_overrides=TARGET_WEIGHTING_COLUMNS)
    else:
        weightings = None

    if elusive_edges.height > 10**4:
        min_cluster_targets = 10
    else:
        min_cluster_targets = 1

    clusters = pipeline(
        elusive_edges,
        read_size,
        weightings,
        anchor_samples=anchor_samples,
        MAX_COASSEMBLY_SIZE=MAX_COASSEMBLY_SIZE,
        MAX_COASSEMBLY_SAMPLES=MAX_COASSEMBLY_SAMPLES,
        MIN_COASSEMBLY_SAMPLES=MIN_COASSEMBLY_SAMPLES,
        MAX_RECOVERY_SAMPLES=MAX_RECOVERY_SAMPLES,
        COASSEMBLY_SAMPLES=COASSEMBLY_SAMPLES,
        EXCLUDE_COASSEMBLIES=EXCLUDE_COASSEMBLIES,
        MIN_CLUSTER_TARGETS=min_cluster_targets,
        MAX_SAMPLES_COMBINATIONS=MAX_SAMPLES_COMBINATIONS,
        single_assembly=single_assembly,
    )
    clusters.write_csv(args.elusive_clusters, separator="\t")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
##########################
### aviary_commands.py ###
##########################
# Author: Samuel Aroney

import os
import polars as pl
from binchicken.binchicken import FAST_AVIARY_MODE
import argparse

# Example shell directive for Snakemake:
# shell:
# """
# python3 binchicken/workflow/scripts/aviary_commands.py \
#   --elusive-clusters {input.elusive_clusters} \
#   --coassemble-commands {output.coassemble_commands} \
#   --recover-commands {output.recover_commands} \
#   --reads-1 {input.reads_1} \
#   --reads-2 {input.reads_2} \
#   --dir {params.dir} \
#   --assemble-threads {params.assemble_threads} \
#   --assemble-memory {params.assemble_memory} \
#   --recover-threads {params.recover_threads} \
#   --recover-memory {params.recover_memory} \
#   --speed {params.speed} \
#   --threads {threads} \
#   --log {log}
# """

def pipeline(coassemblies, reads_1, reads_2, output_dir, assemble_threads, assemble_memory, recover_threads, recover_memory, fast=False):
    output = (
        coassemblies
        .with_columns(
            pl.col("samples").str.split(","),
            pl.col("recover_samples").str.split(","),
            )
        .with_columns(
            coassembly_samples_1 = pl.col("samples").list.eval(pl.element().replace(reads_1)),
            coassembly_samples_2 = pl.col("samples").list.eval(pl.element().replace(reads_2)),
            recover_samples_1 = pl.col("recover_samples").list.eval(pl.element().replace(reads_1)),
            recover_samples_2 = pl.col("recover_samples").list.eval(pl.element().replace(reads_2)),
            )
        .with_columns(
            assemble = pl.concat_str(
                pl.lit("aviary assemble --coassemble -1 "),
                pl.col("coassembly_samples_1").list.join(" "),
                pl.lit(" -2 "),
                pl.col("coassembly_samples_2").list.join(" "),
                pl.lit(" --output "),
                pl.lit(output_dir),
                pl.lit("/coassemble/"),
                pl.col("coassembly"),
                pl.lit("/assemble -n "),
                pl.lit(assemble_threads),
                pl.lit(" -t "),
                pl.lit(assemble_threads),
                pl.lit(" -m "),
                pl.lit(assemble_memory),
                pl.lit(" --skip-qc &> "),
                pl.lit(output_dir),
                pl.lit("/coassemble/logs/"),
                pl.col("coassembly"),
                pl.lit("_assemble.log ")
                ),
            recover = pl.concat_str(
                pl.lit("aviary recover --assembly "),
                pl.lit(output_dir),
                pl.lit("/coassemble/"),
                pl.col("coassembly"),
                pl.lit("/assemble/assembly/final_contigs.fasta -1 "),
                pl.col("recover_samples_1").list.join(" "),
                pl.lit(" -2 "),
                pl.col("recover_samples_2").list.join(" "),
                pl.lit(" --output "),
                pl.lit(output_dir),
                pl.lit("/coassemble/"),
                pl.col("coassembly"),
                pl.lit("/recover"),
                pl.when(pl.lit(fast)).then(pl.lit(" --binning-only --refinery-max-iterations 0")).otherwise(pl.lit("")),
                pl.lit(" -n "),
                pl.lit(recover_threads),
                pl.lit(" -t "),
                pl.lit(recover_threads),
                pl.lit(" -m "),
                pl.lit(recover_memory),
                pl.lit(" --skip-qc &> "),
                pl.lit(output_dir),
                pl.lit("/coassemble/logs/"),
                pl.col("coassembly"),
                pl.lit("_recover.log ")
                ),
            )
    )

    return output.select("assemble", "recover")

def main():
    parser = argparse.ArgumentParser(description="Aviary commands pipeline script.")
    parser.add_argument("--elusive-clusters", required=True, help="Path to elusive clusters input file")
    parser.add_argument("--coassemble-commands", required=True, help="Path to output coassemble commands file")
    parser.add_argument("--recover-commands", required=True, help="Path to output recover commands file")
    parser.add_argument("--reads-1", required=True, help="Named list file of read1")
    parser.add_argument("--reads-2", required=True, help="Named list file of read2")
    parser.add_argument("--dir", required=True, help="Output directory")
    parser.add_argument("--assemble-threads", type=int, required=True, help="Threads for assembly")
    parser.add_argument("--assemble-memory", required=True, help="Memory for assembly")
    parser.add_argument("--recover-threads", type=int, required=True, help="Threads for recovery")
    parser.add_argument("--recover-memory", required=True, help="Memory for recovery")
    parser.add_argument("--speed", default=None, help="Speed mode (FAST_AVIARY_MODE)")
    parser.add_argument("--threads", type=int, default=1, help="Number of threads for Polars")
    parser.add_argument("--log", default=None, help="Log file path")
    args = parser.parse_args()

    os.environ["POLARS_MAX_THREADS"] = str(args.threads)
    import polars as pl

    if args.log:
        import logging
        logging.basicConfig(
            filename=args.log,
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )
    else:
        import logging
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )

    coassemblies = pl.read_csv(args.elusive_clusters, separator="\t")
    if coassemblies.height == 0:
        with open(args.coassemble_commands, "w") as f:
            pass
        with open(args.recover_commands, "w") as f:
            pass
        print("No coassemblies to perform")
        exit(0)

    fast = args.speed == FAST_AVIARY_MODE

    reads_1 = {}
    with open(args.reads_1, "r") as f:
        for line in f:
            sample, read1 = line.strip().split("\t")
            reads_1[sample] = read1
    reads_2 = {}
    with open(args.reads_2, "r") as f:
        for line in f:
            sample, read2 = line.strip().split("\t")
            reads_2[sample] = read2

    coassemblies = pipeline(
        coassemblies,
        reads_1=reads_1,
        reads_2=reads_2,
        output_dir=args.dir,
        assemble_threads=args.assemble_threads,
        assemble_memory=args.assemble_memory,
        recover_threads=args.recover_threads,
        recover_memory=args.recover_memory,
        fast=fast,
    )

    coassemblies.select("assemble").write_csv(args.coassemble_commands, separator="\t", include_header=False)
    coassemblies.select("recover").write_csv(args.recover_commands, separator="\t", include_header=False)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#################################
### summarise_coassemblies.py ###
#################################
# Author: Samuel Aroney

import os
import polars as pl
import argparse

def processing(elusive_clusters, read_size):
    print(f"Polars using {str(pl.thread_pool_size())} threads")

    summary = (
        elusive_clusters
        .select("coassembly", "samples", "length", "total_targets", "total_size")
    )

    if read_size is not None:
        summary = (
            summary
            .with_columns(sample = pl.col("samples").str.split(","))
            .explode("sample")
            .join(read_size, on="sample", how="left", coalesce=True)
            .group_by("coassembly", "samples", "length", "total_targets", "total_size")
            .agg(unmapped_size = pl.sum("read_size"))
        )

    return summary

# Example shell directive for Snakemake:
# shell:
# """
# python3 binchicken/workflow/scripts/summarise_coassemblies.py \
#   --elusive-clusters {input.elusive_clusters} \
#   --read-size {input.read_size} \
#   --summary {output.summary} \
#   --threads {threads}
# """

def main():
    parser = argparse.ArgumentParser(description="Summarise coassemblies pipeline script.")
    parser.add_argument("--elusive-clusters", required=True, help="Path to elusive clusters input file")
    parser.add_argument("--read-size", nargs='?', const=None, default=None, help="Path to read size input file (optional)")
    parser.add_argument("--summary", required=True, help="Path to output summary file")
    parser.add_argument("--threads", type=int, default=1, help="Number of threads for Polars")
    args = parser.parse_args()

    os.environ["POLARS_MAX_THREADS"] = str(args.threads)
    import polars as pl

    elusive_clusters = pl.read_csv(args.elusive_clusters, separator="\t")
    if args.read_size:
        read_size = pl.read_csv(args.read_size, has_header=False, new_columns=["sample", "read_size"])
    else:
        read_size = None

    summary = processing(elusive_clusters, read_size)
    summary.write_csv(args.summary, separator="\t")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#################################
### collect_reference_bins.py ###
#################################
# Author: Samuel Aroney

import polars as pl
import os
import numpy as np
import extern
import argparse

# Example shell directive for Snakemake:
# shell:
# """
# python3 binchicken/workflow/scripts/collect_reference_bins.py \
#   --appraise-binned {input.appraise_binned} \
#   --appraise-unbinned {input.appraise_unbinned} \
#   --genomes {input.genomes} \
#   --sample {params.sample} \
#   --min-appraised {params.min_appraised} \
#   --read {wildcards.read} \
#   --output {output}
# """

def trimmed_mean(data, trim=0.1):
    cut = int(np.floor(len(data) * trim))
    if cut == 0:
        return np.mean(data)
    else:
        a = sorted(data)
        return np.mean(a[cut:-cut])

def pipeline(appraise_binned, appraise_unbinned, sample, MIN_APPRAISED=0.1, TRIM_FRACTION=0.1):
    print(f"Polars using {str(pl.thread_pool_size())} threads")

    appraise_binned = (
        appraise_binned
        .with_columns(pl.col("sample").cast(str))
        .filter(pl.col("sample") == sample)
    )

    appraise_unbinned = (
        appraise_unbinned
        .with_columns(pl.col("sample").cast(str))
        .filter(pl.col("sample") == sample)
    )

    num_binned = sum(appraise_binned.get_column("num_hits").to_list())
    num_unbinned = sum(appraise_unbinned.get_column("num_hits").to_list())
    try:
        perc_binned = num_binned / (num_binned + num_unbinned)
    except ZeroDivisionError:
        return set()

    if perc_binned < MIN_APPRAISED:
        return set()

    reference_bins = (
        appraise_binned
        .with_columns(pl.col("found_in").str.split(","))
        .explode("found_in")
        .with_columns(pl.col("found_in").str.replace("_protein$", ""))
        .group_by(["gene", "found_in"])
        .agg(pl.col("coverage").sum())
        .pivot(values="coverage", index="gene", on="found_in", aggregate_function=None)
        .unpivot(index="gene", variable_name="found_in", value_name="coverage")
        .fill_null(0)
        .group_by("found_in")
        .agg(
            (pl.col("coverage").len() * TRIM_FRACTION).floor().cast(int).alias("cut"),
            pl.col("coverage")
            )
        .with_columns(
            pl.col("coverage")
                .list.sort()
                .list.slice(pl.col("cut"), pl.col("coverage").list.len() - 2 * pl.col("cut"))
                .list.mean()
            )
        .filter(pl.col("coverage") > 0)
        .get_column("found_in")
        .to_list()
    )

    return set(reference_bins)

def main():
    parser = argparse.ArgumentParser(description="Collect reference bins pipeline script.")
    parser.add_argument("--appraise-binned", required=True, help="Path to appraise binned input file")
    parser.add_argument("--appraise-unbinned", required=True, help="Path to appraise unbinned input file")
    parser.add_argument("--genomes", required=True, help="Named list file of genome fasta files (indexed by bin name)")
    parser.add_argument("--sample", required=True, help="Sample name")
    parser.add_argument("--min-appraised", type=float, default=0.1, help="Minimum appraised fraction")
    parser.add_argument("--read", required=True, help="Read wildcard value")
    parser.add_argument("--output", required=True, help="Output fasta file path")
    parser.add_argument("--threads", type=int, default=1, help="Number of threads for Polars")
    args = parser.parse_args()

    os.environ["POLARS_MAX_THREADS"] = str(args.threads)
    import polars as pl

    appraise_binned = pl.read_csv(args.appraise_binned, separator="\t")
    appraise_unbinned = pl.read_csv(args.appraise_unbinned, separator="\t")

    # Build a mapping from bin name to genome file path
    genomes_dict = {}
    with open(args.genomes, "r") as f:
        for line in f:
            bin_name, genome_path = line.strip().split("\t")
            genomes_dict[bin_name] = genome_path

    reference_bins = pipeline(appraise_binned, appraise_unbinned, args.sample, MIN_APPRAISED=args.min_appraised)

    if len(reference_bins) == 0:
        print(f"Warning: No reference bins found for {args.read}")
        cmd = f"touch {args.output}"
        extern.run(cmd)
    else:
        with open(str(args.output), "w") as outfile:
            for bin in reference_bins:
                bin_name = os.path.splitext(os.path.basename(genomes_dict[bin]))[0]
                with open(genomes_dict[bin], "r") as infile:
                    for line in infile:
                        if line.startswith(">"):
                            line = f">{bin_name}~{line[1:]}"
                        outfile.write(line)

if __name__ == "__main__":
    main()



import os
import importlib.resources
import re
import ast

def fasta_to_name(filename):
    sample_name = os.path.basename(filename)
    # Put .gz first so it is stripped off first.
    for extension in (".gz",".fna",".fq",".fastq",".fasta",".fa"):
        if sample_name.endswith(extension):
            sample_name = sample_name[0:(len(sample_name)-len(extension))]
            if extension != ".gz":
                break
    return sample_name

def parse_snake_dict(s):
    # Remove leading/trailing whitespace
    s = s.strip()
    # If it doesn't start with '{', it's not a dict
    if not s.startswith("{"):
        raise ValueError("Input does not look like a dict: " + s)
    # Add quotes around keys and values if missing
    # This regex finds keys and values that are not quoted and adds quotes
    s = re.sub(r'([{,]\s*)([^:,}{]+)\s*:', r'\1"\2":', s)
    s = re.sub(r':\s*([^,}{]+)', lambda m: ': "' + m.group(1).strip().strip('"\'') + '"', s)
    # Now safe to use ast.literal_eval
    return ast.literal_eval(s)

def pixi_run_func():
    with importlib.resources.path("binchicken", "pixi.toml") as manifest_path:
        return f"pixi run --frozen --manifest-path {manifest_path}"

pixi_run = pixi_run_func()



#!/usr/bin/env python3
##############################
### abundance_weighting.py ###
##############################
# Author: Samuel Aroney

import os
import polars as pl
import logging
import argparse

APPRAISE_COLUMNS = {
    "gene": str,
    "sample": str,
    "sequence": str,
    "num_hits": int,
    "coverage": float,
    "taxonomy": str,
    "found_in": str,
}

WEIGHTING_COLUMNS = {
    "gene": str,
    "sequence": str,
    "weight": float,
}

# Example shell directive for Snakemake:
# shell:
# """
# python3 binchicken/workflow/scripts/abundance_weighting.py \
#   --unbinned {input.unbinned} \
#   --binned {input.binned} \
#   --weighted {output.weighted} \
#   --samples {input.samples} \
#   --threads {threads} \
#   --log {log}
# """

def pipeline(unbinned, binned, samples=None):
    logging.info(f"Polars using {str(pl.thread_pool_size())} threads")

    if len(unbinned) == 0:
        logging.warning("No unbinned sequences found")
        return pl.DataFrame(schema=WEIGHTING_COLUMNS)

    total_coverage = (
        pl.concat([binned, unbinned])
        .group_by("sample", "gene")
        .agg(total_coverage = pl.sum("coverage"))
    )

    if samples:
        total_coverage = (
            total_coverage
            .filter(pl.col("sample").is_in(samples))
        )

    gene_sequences = (
        unbinned
        .select("gene", "sequence")
        .unique()
    )

    weighted = (
        total_coverage
        .join(gene_sequences, on=["gene"])
        .join(unbinned, on=["sample", "gene", "sequence"], how="full")
        .filter(pl.col("sample").is_not_null())
        .with_columns(pl.col("coverage").fill_null(0))
        .with_columns(weight = pl.col("coverage") / pl.col("total_coverage"))
        .group_by("gene", "sequence")
        .agg(pl.mean("weight"))
        .filter(pl.col("weight") > 0)
    )

    return weighted

def main():
    parser = argparse.ArgumentParser(description="Abundance weighting pipeline script.")
    parser.add_argument("--unbinned", required=True, help="Path to unbinned input file")
    parser.add_argument("--binned", required=True, help="Path to binned input file")
    parser.add_argument("--weighted", required=True, help="Path to output weighted file")
    parser.add_argument("--samples", default=None, help="List file of sample names")
    parser.add_argument("--threads", type=int, default=1, help="Number of threads for Polars")
    parser.add_argument("--log", default=None, help="Log file path")
    args = parser.parse_args()

    os.environ["POLARS_MAX_THREADS"] = str(args.threads)
    import polars as pl

    if args.log:
        logging.basicConfig(
            filename=args.log,
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )
    else:
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )

    unbinned = pl.read_csv(args.unbinned, separator="\t", schema_overrides=APPRAISE_COLUMNS)
    binned = pl.read_csv(args.binned, separator="\t", schema_overrides=APPRAISE_COLUMNS)

    if not args.samples:
        samples = None
    elif os.path.getsize(args.samples) == 0:
        samples = None
    else:
        samples = pl.read_csv(args.samples, has_header=False, new_columns=["sample"]).get_column("sample").to_list()

    weighted = pipeline(unbinned, binned, samples)
    weighted.write_csv(args.weighted, separator="\t")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#########################
### sketch_samples.py ###
#########################
# Author: Samuel Aroney

import polars as pl
import os
import logging
from sourmash import MinHash, SourmashSignature
from sourmash.sourmash_args import SaveSignaturesToLocation
from concurrent.futures import ProcessPoolExecutor
import extern
import re
import argparse

SINGLEM_OTU_TABLE_SCHEMA = {
    "gene": str,
    "sample": str,
    "sequence": str,
    "num_hits": int,
    "coverage": float,
    "taxonomy": str,
    }

# Example shell directive for Snakemake:
# shell:
# """
# python3 binchicken/workflow/scripts/sketch_samples.py \
#   --unbinned {input.unbinned} \
#   --sketch {output.sketch} \
#   --taxa-of-interest {params.taxa_of_interest} \
#   --threads {threads} \
#   --log {log}
# """

def process_groups(groups, output_path):
    with SaveSignaturesToLocation(output_path) as save_sigs:
        for group in groups:
            sample, sequences = group
            mh = MinHash(n=0, ksize=60, scaled=1, track_abundance=False)
            for seq in sequences:
                mh.add_sequence(seq.replace("-", "A").replace("N", "A"))
            signature = SourmashSignature(mh, name=sample)
            save_sigs.add(signature)

def processing(unbinned_path, output_path, taxa_of_interest=None, threads=1, samples_per_group=1000):
    output_dir = os.path.dirname(output_path)

    with open(unbinned_path) as f:
        logging.info(f"Reading unbinned OTU table from {unbinned_path}")
        if taxa_of_interest:
            logging.info(f"Filtering for taxa of interest: {taxa_of_interest}")

        logging.info("Generating sketches in separate threads")
        with ProcessPoolExecutor(max_workers=threads) as executor:
            current_sample = ""
            current_sequences = []
            i = 0
            group_subset = []
            futures = []
            for line in f:
                line = line.strip()
                if line == "gene\tsample\tsequence\tnum_hits\tcoverage\ttaxonomy":
                    continue

                sample = line.split("\t")[1]
                sequence = line.split("\t")[2]

                if taxa_of_interest:
                    taxonomy = line.split("\t")[5]
                    if not re.search(taxa_of_interest, taxonomy):
                        continue

                if not current_sample:
                    current_sample = sample

                if sample != current_sample:
                    group_subset.append((current_sample, current_sequences))
                    current_sample = sample
                    current_sequences = [sequence]

                    if len(group_subset) == samples_per_group:
                        output_subpath = os.path.join(output_dir, f"signatures_thread_{i}.sig")
                        future = executor.submit(process_groups, group_subset, output_subpath)
                        futures.append(future)
                        i += 1
                        group_subset = []
                else:
                    current_sequences.append(sequence)

            # Submit any remaining groups that are smaller than samples_per_group
            if sample == current_sample:
                group_subset.append((current_sample, current_sequences))

            if group_subset:
                output_subpath = os.path.join(output_dir, f"signatures_thread_{i}.sig")
                future = executor.submit(process_groups, group_subset, output_subpath)
                futures.append(future)

            for future in futures:
                future.result()

    logging.info("Concatenating sketches")
    extern.run(f"sourmash sig cat {os.path.join(output_dir, 'signatures_thread_*.sig')} -o {output_path}")

    logging.info("Done")
    return output_path

def main():
    parser = argparse.ArgumentParser(description="Sketch samples pipeline script.")
    parser.add_argument("--unbinned", required=True, help="Path to unbinned input file")
    parser.add_argument("--sketch", required=True, help="Path to output sketch file")
    parser.add_argument("--taxa-of-interest", nargs='?', const=None, default=None, help="Taxa of interest string")
    parser.add_argument("--threads", type=int, default=1, help="Number of threads for Polars and sketching")
    parser.add_argument("--log", default=None, help="Log file path")
    args = parser.parse_args()

    os.environ["POLARS_MAX_THREADS"] = str(args.threads)
    import polars as pl

    if args.log:
        logging.basicConfig(
            filename=args.log,
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )
    else:
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )
    logging.info(f"Polars using {str(pl.thread_pool_size())} threads")

    processing(
        args.unbinned,
        args.sketch,
        taxa_of_interest=args.taxa_of_interest,
        threads=args.threads
    )

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import unittest
import os
from bird_tool_utils import in_tempdir
import extern
from snakemake.io import load_configfile

path_to_data = os.path.join(os.path.dirname(os.path.realpath(__file__)),'data')

METAPACKAGE = os.path.join(path_to_data, "singlem_metapackage.smpkg")
MOCK_COASSEMBLE = os.path.join(path_to_data, "mock_coassemble")
MOCK_UNBINNED = os.path.join(MOCK_COASSEMBLE, "coassemble", "appraise", "unbinned.otu_table.tsv")
MOCK_BINNED = os.path.join(MOCK_COASSEMBLE, "coassemble", "appraise", "binned.otu_table.tsv")
MOCK_TARGETS = os.path.join(MOCK_COASSEMBLE, "coassemble", "target", "targets.tsv")
MOCK_ELUSIVE_EDGES = os.path.join(MOCK_COASSEMBLE, "coassemble", "target", "elusive_edges.tsv")
MOCK_ELUSIVE_CLUSTERS = os.path.join(MOCK_COASSEMBLE, "coassemble", "target", "elusive_clusters.tsv")
MOCK_SUMMARY = os.path.join(MOCK_COASSEMBLE, "coassemble", "summary.tsv")

MOCK_COASSEMBLIES = ' '.join([os.path.join(MOCK_COASSEMBLE, "coassemble", "coassemble", "coassembly_0")])
MOCK_GENOMES = " ".join([
    os.path.join(MOCK_COASSEMBLE, "coassemble", "coassemble", "coassembly_0", "recover", "bins", "final_bins", "bin_1.fna"),
    os.path.join(MOCK_COASSEMBLE, "coassemble", "coassemble", "coassembly_0", "recover", "bins", "final_bins", "bin_2.fna"),
    os.path.join(MOCK_COASSEMBLE, "coassemble", "coassemble", "coassembly_0", "recover", "bins", "final_bins", "bin_3.fna"),
])

GENOMES = " ".join([os.path.join(path_to_data, "GB_GCA_013286235.1.fna")])
TWO_GENOMES = " ".join([
    os.path.join(path_to_data, "GB_GCA_013286235.1.fna"),
    os.path.join(path_to_data, "GB_GCA_013286235.2.fna"),
    ])

def write_string_to_file(string, filename):
    with open(filename, "w") as f:
        f.write("\n".join(string.split(" ")))

class Tests(unittest.TestCase):
    def test_evaluate(self):
        with in_tempdir():
            cmd = (
                f"binchicken evaluate "
                f"--coassemble-output {MOCK_COASSEMBLE} "
                f"--aviary-outputs {MOCK_COASSEMBLIES} "
                f"--checkm-version 2 "
                f"--singlem-metapackage {METAPACKAGE} "
                f"--output test "
            )
            extern.run(cmd)

            config_path = os.path.join("test", "config.yaml")
            self.assertTrue(os.path.exists(config_path))

            summary_path = os.path.join("test", "evaluate", "evaluate", "summary_stats.tsv")
            self.assertTrue(os.path.exists(summary_path))
            expected = "\n".join(
                [
                    "\t".join([
                        "coassembly",
                        "statistic",
                        "within",
                        "match",
                        "nonmatch",
                        "total",
                        "match_percent",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "sequences",
                        "targets",
                        "1",
                        "1",
                        "2",
                        "50.0",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "taxonomy",
                        "targets",
                        "1",
                        "1",
                        "2",
                        "50.0",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "bins",
                        "recovery",
                        "1",
                        "1",
                        "2",
                        "50.0",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "nontarget_bin_sequences",
                        "recovery",
                        "1",
                        "2",
                        "3",
                        "33.33",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "nontarget_unbin_sequences",
                        "recovery",
                        "0",
                        "3",
                        "3",
                        "0.0",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "novel_sequences",
                        "recovery",
                        "1",
                        "2",
                        "3",
                        "33.33",
                    ]),
                    ""
                ]
            )
            with open(summary_path) as f:
                self.assertEqual(expected, f.read())

            summarise_path = os.path.join("test", "evaluate", "summarise", "bins_summarised.otu_table.tsv")
            self.assertTrue(os.path.exists(summarise_path))
            with open(summarise_path) as f:
                observed = f.read()

            otu1 = "\t".join(["S3.7.ribosomal_protein_S7", "coassembly_0-0_transcripts", "TATCAAGTTCCACAAGAAGTTAGAGGAGAAAGAAGAATCTCGTTAGCTATTAGATGGATT", "1", "1.14", "Root; d__Bacteria"])
            otu2 = "\t".join(["S3.7.ribosomal_protein_S7", "coassembly_0-1_transcripts", "TACCAGGTTCCTGTTGAAGTTCGCCCATCACGCCGTTCTGCTTTGGCAATGCGCTGGGTG", "1", "1.14", "Root; d__Bacteria; p__Proteobacteria; c__Gammaproteobacteria; o__Burkholderiales; f__Burkholderiaceae; g__Polynucleobacter; s__Polynucleobacter_sp018688335"])
            otu3 = "\t".join(["S3.7.ribosomal_protein_S7", "coassembly_0-1_transcripts", "TTCCAGGTGCCTACTGAAGTTCGTCCCGAGCGTAAAATTGCATTGGGTATGAAATGGCTG", "1", "1.18", "Root; d__Bacteria; p__Bacteroidota; c__Bacteroidia; o__Sphingobacteriales; f__Sphingobacteriaceae; g__Mucilaginibacter; s__Mucilaginibacter_sp013286235"])

            self.assertTrue(otu1 in observed)
            self.assertTrue(otu2 in observed)
            self.assertTrue(otu3 in observed)

    def test_evaluate_specified_files(self):
        with in_tempdir():
            cmd = (
                f"binchicken evaluate "
                f"--coassemble-unbinned {MOCK_UNBINNED} "
                f"--coassemble-binned {MOCK_BINNED} "
                f"--coassemble-targets {MOCK_TARGETS} "
                f"--coassemble-elusive-edges {MOCK_ELUSIVE_EDGES} "
                f"--coassemble-elusive-clusters {MOCK_ELUSIVE_CLUSTERS} "
                f"--coassemble-summary {MOCK_SUMMARY} "
                f"--aviary-outputs {MOCK_COASSEMBLIES} "
                f"--singlem-metapackage {METAPACKAGE} "
                f"--output test "
                f"--dryrun "
                f"--snakemake-args \" --quiet\" "
            )
            output = extern.run(cmd)

            self.assertTrue("prodigal_bins" in output)
            self.assertTrue("singlem_pipe_bins" in output)
            self.assertTrue("singlem_summarise_bins" in output)
            self.assertTrue("cluster_original_bins" not in output)
            self.assertTrue("cluster_updated_bins" not in output)
            self.assertTrue("summarise_clusters" not in output)
            self.assertTrue("evaluate" in output)
            self.assertTrue("evaluate_plots" in output)

    def test_evaluate_genome_input(self):
        with in_tempdir():
            cmd = (
                f"binchicken evaluate "
                f"--coassemble-output {MOCK_COASSEMBLE} "
                f"--new-genomes {MOCK_GENOMES} "
                f"--prodigal-meta "
                f"--coassembly-run coassembly_0 "
                f"--singlem-metapackage {METAPACKAGE} "
                f"--output test "
            )
            extern.run(cmd)

            config_path = os.path.join("test", "config.yaml")
            self.assertTrue(os.path.exists(config_path))

            summary_path = os.path.join("test", "evaluate", "evaluate", "summary_stats.tsv")
            self.assertTrue(os.path.exists(summary_path))
            expected = "\n".join(
                [
                    "\t".join([
                        "coassembly",
                        "statistic",
                        "within",
                        "match",
                        "nonmatch",
                        "total",
                        "match_percent",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "sequences",
                        "targets",
                        "1",
                        "1",
                        "2",
                        "50.0",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "taxonomy",
                        "targets",
                        "1",
                        "1",
                        "2",
                        "50.0",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "bins",
                        "recovery",
                        "1",
                        "2",
                        "3",
                        "33.33",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "nontarget_bin_sequences",
                        "recovery",
                        "1",
                        "2",
                        "3",
                        "33.33",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "nontarget_unbin_sequences",
                        "recovery",
                        "0",
                        "3",
                        "3",
                        "0.0",
                    ]),
                    "\t".join([
                        "coassembly_0",
                        "novel_sequences",
                        "recovery",
                        "1",
                        "2",
                        "3",
                        "33.33",
                    ]),
                    ""
                ]
            )
            with open(summary_path) as f:
                self.assertEqual(expected, f.read())

            summarise_path = os.path.join("test", "evaluate", "summarise", "bins_summarised.otu_table.tsv")
            self.assertTrue(os.path.exists(summarise_path))
            with open(summarise_path) as f:
                observed = f.read()

            otu1 = "\t".join(["S3.7.ribosomal_protein_S7", "coassembly_0-bin_1_transcripts", "TATCAAGTTCCACAAGAAGTTAGAGGAGAAAGAAGAATCTCGTTAGCTATTAGATGGATT", "1", "1.14", "Root; d__Bacteria"])
            otu2 = "\t".join(["S3.7.ribosomal_protein_S7", "coassembly_0-bin_3_transcripts", "TACCAGGTTCCTGTTGAAGTTCGCCCATCACGCCGTTCTGCTTTGGCAATGCGCTGGGTG", "1", "1.14", "Root; d__Bacteria; p__Proteobacteria; c__Gammaproteobacteria; o__Burkholderiales; f__Burkholderiaceae; g__Polynucleobacter; s__Polynucleobacter_sp018688335"])
            otu3 = "\t".join(["S3.7.ribosomal_protein_S7", "coassembly_0-bin_3_transcripts", "TTCCAGGTGCCTACTGAAGTTCGTCCCGAGCGTAAAATTGCATTGGGTATGAAATGGCTG", "1", "1.14", "Root; d__Bacteria; p__Bacteroidota; c__Bacteroidia; o__Sphingobacteriales; f__Sphingobacteriaceae; g__Mucilaginibacter; s__Mucilaginibacter_sp013286235"])

            self.assertTrue(otu1 in observed)
            self.assertTrue(otu2 in observed)
            self.assertTrue(otu3 in observed)

    def test_evaluate_genome_files_input(self):
        with in_tempdir():
            write_string_to_file(MOCK_GENOMES, "genomes")

            cmd = (
                f"binchicken evaluate "
                f"--coassemble-output {MOCK_COASSEMBLE} "
                f"--new-genomes-list genomes "
                f"--prodigal-meta "
                f"--coassembly-run coassembly_0 "
                f"--singlem-metapackage {METAPACKAGE} "
                f"--output test "
                f"--dryrun "
                f"--snakemake-args \" --quiet\" "
            )
            output = extern.run(cmd)

            self.assertTrue("prodigal_bins" in output)
            self.assertTrue("singlem_pipe_bins" in output)
            self.assertTrue("singlem_summarise_bins" in output)
            self.assertTrue("cluster_original_bins" not in output)
            self.assertTrue("cluster_updated_bins" not in output)
            self.assertTrue("summarise_clusters" not in output)
            self.assertTrue("evaluate" in output)
            self.assertTrue("evaluate_plots" in output)

    def test_evaluate_default_config(self):
        with in_tempdir():
            cmd = (
                f"binchicken evaluate "
                f"--coassemble-output {MOCK_COASSEMBLE} "
                f"--aviary-outputs {MOCK_COASSEMBLIES} "
                f"--singlem-metapackage {METAPACKAGE} "
                f"--output test "
                f"--dryrun "
            )
            extern.run(cmd)

            config = load_configfile(os.path.join("test", "config.yaml"))
            self.assertEqual(config["max_threads"], 8)
            self.assertEqual(config["checkm_version"], 2)
            self.assertEqual(config["min_completeness"], 70)
            self.assertEqual(config["max_contamination"], 10)

    def test_evaluate_metapackage_env_variable(self):
        with in_tempdir():
            os.environ['SINGLEM_METAPACKAGE_PATH'] = METAPACKAGE
            cmd = (
                f"binchicken evaluate "
                f"--coassemble-output {MOCK_COASSEMBLE} "
                f"--aviary-outputs {MOCK_COASSEMBLIES} "
                f"--output test "
                f"--snakemake-args \"singlem_summarise_bins\" "
            )
            import subprocess
            _ = subprocess.run(cmd, shell=True, check=True, capture_output=True, env=os.environ)

            summarise_path = os.path.join("test", "evaluate", "summarise", "bins_summarised.otu_table.tsv")
            self.assertTrue(os.path.exists(summarise_path))
            with open(summarise_path) as f:
                observed = f.read()

            otu1 = "\t".join(["S3.7.ribosomal_protein_S7", "coassembly_0-0_transcripts", "TATCAAGTTCCACAAGAAGTTAGAGGAGAAAGAAGAATCTCGTTAGCTATTAGATGGATT", "1", "1.14", "Root; d__Bacteria"])
            otu2 = "\t".join(["S3.7.ribosomal_protein_S7", "coassembly_0-1_transcripts", "TACCAGGTTCCTGTTGAAGTTCGCCCATCACGCCGTTCTGCTTTGGCAATGCGCTGGGTG", "1", "1.14", "Root; d__Bacteria; p__Proteobacteria; c__Gammaproteobacteria; o__Burkholderiales; f__Burkholderiaceae; g__Polynucleobacter; s__Polynucleobacter_sp018688335"])
            otu3 = "\t".join(["S3.7.ribosomal_protein_S7", "coassembly_0-1_transcripts", "TTCCAGGTGCCTACTGAAGTTCGTCCCGAGCGTAAAATTGCATTGGGTATGAAATGGCTG", "1", "1.18", "Root; d__Bacteria; p__Bacteroidota; c__Bacteroidia; o__Sphingobacteriales; f__Sphingobacteriaceae; g__Mucilaginibacter; s__Mucilaginibacter_sp013286235"])

            self.assertTrue(otu1 in observed)
            self.assertTrue(otu2 in observed)
            self.assertTrue(otu3 in observed)

    def test_evaluate_cluster(self):
        with in_tempdir():
            cmd = (
                f"binchicken evaluate "
                f"--coassemble-output {MOCK_COASSEMBLE} "
                f"--aviary-outputs {MOCK_COASSEMBLIES} "
                f"--cluster "
                f"--genomes {TWO_GENOMES} "
                f"--checkm-version 2 "
                f"--singlem-metapackage {METAPACKAGE} "
                f"--output test "
            )
            extern.run(cmd)

            config_path = os.path.join("test", "config.yaml")
            self.assertTrue(os.path.exists(config_path))

            summary_path = os.path.join("test", "evaluate", "evaluate", "summary_stats.tsv")
            self.assertTrue(os.path.exists(summary_path))

            cluster_path = os.path.join("test", "evaluate", "evaluate", "cluster_stats.csv")
            self.assertTrue(os.path.exists(cluster_path))
            expected = "\n".join(
                [
                    ",".join([
                        "original",
                        "2",
                    ]),
                    ",".join([
                        "coassembly_0",
                        "4",
                    ]),
                    ""
                ]
            )
            with open(cluster_path) as f:
                self.assertEqual(expected, f.read())


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#####################
### no_genomes.py ###
#####################
# Author: Samuel Aroney

import polars as pl
import os
import argparse

SINGLEM_OTU_TABLE_SCHEMA = {
    "gene": str,
    "sample": str,
    "sequence": str,
    "num_hits": int,
    "coverage": float,
    "taxonomy": str,
    }

# Example shell directive for Snakemake:
# shell:
# """
# python3 binchicken/workflow/scripts/no_genomes.py \
#   --reads {input.reads} \
#   --binned {output.binned} \
#   --unbinned {output.unbinned} \
#   --threads {threads}
# """

def processing(reads):
    print(f"Polars using {str(pl.thread_pool_size())} threads")

    unbinned = (
        reads
        .with_columns(found_in = pl.lit(""))
        .filter(pl.col("gene") != "S3.18.EIF_2_alpha")
    )
    binned = unbinned.filter(False)

    return(binned, unbinned)

def main():
    parser = argparse.ArgumentParser(description="No genomes pipeline script.")
    parser.add_argument("--reads", required=True, help="List file of input read files")
    parser.add_argument("--binned", required=True, help="Path to output binned file")
    parser.add_argument("--unbinned", required=True, help="Path to output unbinned file")
    parser.add_argument("--threads", type=int, default=1, help="Number of threads for Polars")
    args = parser.parse_args()

    os.environ["POLARS_MAX_THREADS"] = str(args.threads)
    import polars as pl

    reads = pl.scan_csv(args.reads, separator="\t", schema_overrides=SINGLEM_OTU_TABLE_SCHEMA)
    binned, unbinned = processing(reads)

    binned.sink_csv(args.binned, separator="\t")
    unbinned.sink_csv(args.unbinned, separator="\t")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
###########################
### target_weighting.py ###
###########################
# Author: Samuel Aroney

import os
import polars as pl
import logging
import argparse

TARGET_COLUMNS = {
    "gene": str,
    "sample": str,
    "sequence": str,
    "num_hits": int,
    "coverage": float,
    "taxonomy": str,
    "found_in": str,
    "target": str,
}

WEIGHTING_COLUMNS = {
    "gene": str,
    "sequence": str,
    "weight": float,
}

TARGET_WEIGHTING_COLUMNS = {
    "target": str,
    "weight": float,
}

# Example shell directive for Snakemake:
# shell:
# """
# python3 binchicken/workflow/scripts/target_weighting.py \
#   --targets {input.targets} \
#   --weighting {input.weighting} \
#   --targets-weighted {output.targets_weighted} \
#   --threads {threads} \
#   --log {log}
# """

def pipeline(targets, weighting):
    logging.info(f"Polars using {str(pl.thread_pool_size())} threads")

    if len(targets) == 0:
        logging.warning("No targets found")
        return pl.DataFrame(schema=TARGET_WEIGHTING_COLUMNS)

    weighted = (
        targets
        .select("gene", "sequence", "target")
        .unique()
        .join(weighting, on=["gene", "sequence"])
        .drop("gene", "sequence")
    )

    return weighted

def main():
    parser = argparse.ArgumentParser(description="Target weighting pipeline script.")
    parser.add_argument("--targets", required=True, help="Path to input targets file")
    parser.add_argument("--weighting", required=True, help="Path to input weighting file")
    parser.add_argument("--targets-weighted", required=True, help="Path to output targets weighted file")
    parser.add_argument("--threads", type=int, default=1, help="Number of threads for Polars")
    parser.add_argument("--log", default=None, help="Log file path")
    args = parser.parse_args()

    os.environ["POLARS_MAX_THREADS"] = str(args.threads)
    import polars as pl

    if args.log:
        logging.basicConfig(
            filename=args.log,
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )
    else:
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y/%m/%d %I:%M:%S %p'
        )

    targets = pl.read_csv(args.targets, separator="\t", schema_overrides=TARGET_COLUMNS)
    weighting = pl.read_csv(args.weighting, separator="\t", schema_overrides=WEIGHTING_COLUMNS)
    targets_weighted = pipeline(targets, weighting)
    targets_weighted.write_csv(args.targets_weighted, separator="\t")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import unittest
from binchicken.common import fasta_to_name, parse_snake_dict

class Tests(unittest.TestCase):
    def test_fasta_to_name(self):
        filename1 = "data/sample.1.fq.gz"

        expected = "sample.1"
        observed = fasta_to_name(filename1)
        self.assertEqual(expected, observed)

    def test_fasta_to_name_R(self):
        filename1 = "data/sample_R1.fq.gz"

        expected = "sample_R1"
        observed = fasta_to_name(filename1)
        self.assertEqual(expected, observed)

    def test_fasta_to_name_1(self):
        filename1 = "data/sample_1.fq.gz"

        expected = "sample_1"
        observed = fasta_to_name(filename1)
        self.assertEqual(expected, observed)

    def test_fasta_to_name_fna_gz(self):
        filename1 = "data/sample_1.fna.gz"

        expected = "sample_1"
        observed = fasta_to_name(filename1)
        self.assertEqual(expected, observed)

    def test_fasta_to_name_fna(self):
        filename1 = "data/sample_1.fna"

        expected = "sample_1"
        observed = fasta_to_name(filename1)
        self.assertEqual(expected, observed)

    def test_fasta_to_name_fa(self):
        filename1 = "data/sample_1.fa"

        expected = "sample_1"
        observed = fasta_to_name(filename1)
        self.assertEqual(expected, observed)

    def test_fasta_to_name_fasta(self):
        filename1 = "data/sample_1.fasta"

        expected = "sample_1"
        observed = fasta_to_name(filename1)
        self.assertEqual(expected, observed)

    def test_parse_snake_dict(self):
        snake_dict = '{sample_1: /hpccs01/home/aroneys/src/binchicken/test/data/sample_1.1.fq, sample_2: /hpccs01/home/aroneys/src/binchicken/test/data/sample_2.1.fq, sample_3: /hpccs01/home/aroneys/src/binchicken/test/data/sample_3.1.fq, sample_5: /hpccs01/home/aroneys/src/binchicken/test/data/sample_5.1.fq}'

        expected = {
            "sample_1": "/hpccs01/home/aroneys/src/binchicken/test/data/sample_1.1.fq",
            "sample_2": "/hpccs01/home/aroneys/src/binchicken/test/data/sample_2.1.fq",
            "sample_3": "/hpccs01/home/aroneys/src/binchicken/test/data/sample_3.1.fq",
            "sample_5": "/hpccs01/home/aroneys/src/binchicken/test/data/sample_5.1.fq",
            }
        observed = parse_snake_dict(snake_dict)
        self.assertEqual(expected, observed)

    def test_parse_snake_dict_empty(self):
        snake_dict = '{}'

        expected = {}
        observed = parse_snake_dict(snake_dict)
        self.assertEqual(expected, observed)

    def test_parse_snake_dict_genome(self):
        snake_dict = '{GB_GCA_013286235.1: /home/runner/work/binchicken/binchicken/test/data/GB_GCA_013286235.1.fna}'

        expected = {
            "GB_GCA_013286235.1": "/home/runner/work/binchicken/binchicken/test/data/GB_GCA_013286235.1.fna",
            }
        observed = parse_snake_dict(snake_dict)
        self.assertEqual(expected, observed)

    def test_parse_snake_dict_special(self):
        snake_dict = '{sample-1: /path/to/file-1.fq, sample_2: /path/to/file_2.fq}'

        expected = {
            "sample-1": "/path/to/file-1.fq",
            "sample_2": "/path/to/file_2.fq"
        }
        observed = parse_snake_dict(snake_dict)
        self.assertEqual(expected, observed)

    def test_parse_snake_dict_numeric(self):
        snake_dict = '{1: /path/to/file1.fq, 2: /path/to/file2.fq}'

        expected = {
            "1": "/path/to/file1.fq",
            "2": "/path/to/file2.fq"
        }
        observed = parse_snake_dict(snake_dict)
        self.assertEqual(expected, observed)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import unittest
import os
os.environ["POLARS_MAX_THREADS"] = "8"
import numpy as np
import polars as pl
from polars.testing import assert_frame_equal
from bird_tool_utils import in_tempdir
from binchicken.workflow.scripts.target_elusive import get_clusters, pipeline, streaming_pipeline

SAMPLE_DISTANCES_COLUMNS = {
    "query_name": str,
    "match_name": str,
    "jaccard": float,
}

CLUSTERS_COLUMNS = {
    "samples": str,
}

APPRAISE_COLUMNS = {
    "gene": str,
    "sample": str,
    "sequence": str,
    "num_hits": int,
    "coverage": float,
    "taxonomy": str,
    "found_in": str,
}

TARGETS_COLUMNS = {
    "gene": str,
    "sample": str,
    "sequence": str,
    "num_hits": int,
    "coverage": float,
    "taxonomy": str,
    "target": str,
}

EDGES_COLUMNS = {
    "style": str,
    "cluster_size": int,
    "samples": str,
    "target_ids": str,
}

class Tests(unittest.TestCase):
    def assertDataFrameEqual(self, a, b):
        assert_frame_equal(a, b, check_dtypes=False, check_row_order=False)

    def test_get_clusters(self):
        sample_distances = pl.DataFrame([
            ["sample_1", "sample_2", 1-0.5],
            ["sample_1", "sample_3", 1-1],
            ["sample_2", "sample_3", 1-0.9],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["sample_1", "sample_2", "sample_3"])

        expected_clusters = pl.DataFrame([
            ["sample_1,sample_2"],
            ["sample_2,sample_3"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(sample_distances, samples)
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_suffix(self):
        sample_distances = pl.DataFrame([
            ["sample_1.1", "sample_2.1", 1-0.5],
            ["sample_1.1", "sample_3.1", 1-1],
            ["sample_2.1", "sample_3.1", 1-0.9],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["sample_1", "sample_2", "sample_3"])

        expected_clusters = pl.DataFrame([
            ["sample_1.1,sample_2.1"],
            ["sample_2.1,sample_3.1"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(sample_distances, samples)
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_suffix_underscore(self):
        sample_distances = pl.DataFrame([
            ["sample_1_1", "sample_2_1", 1-0.5],
            ["sample_1_1", "sample_3_1", 1-1],
            ["sample_2_1", "sample_3_1", 1-0.9],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["sample_1_1", "sample_2_1", "sample_3_1"])

        expected_clusters = pl.DataFrame([
            ["sample_1_1,sample_2_1"],
            ["sample_2_1,sample_3_1"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(sample_distances, samples)
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_suffix_R(self):
        sample_distances = pl.DataFrame([
            ["sample_1_R1", "sample_2_R1", 1-0.5],
            ["sample_1_R1", "sample_3_R1", 1-1],
            ["sample_2_R1", "sample_3_R1", 1-0.9],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["sample_1_R1", "sample_2_R1", "sample_3_R1"])

        expected_clusters = pl.DataFrame([
            ["sample_1_R1,sample_2_R1"],
            ["sample_2_R1,sample_3_R1"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(sample_distances, samples)
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_large_precluster_size(self):
        sample_distances = pl.DataFrame([
            ["sample_1", "sample_2", 1-0.5],
            ["sample_1", "sample_3", 1-1],
            ["sample_2", "sample_3", 1-0.9],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["sample_1", "sample_2", "sample_3"])

        expected_clusters = pl.DataFrame([
            ["sample_1,sample_2"],
            ["sample_2,sample_3"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(sample_distances, samples, PRECLUSTER_SIZE=10)
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_size_three(self):
        sample_distances = pl.DataFrame([
            ["1", "2", 1-0.5],
            ["1", "3", 1-1],
            ["2", "3", 1-0.9],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["1", "2", "3"])

        expected_clusters = pl.DataFrame([
            ["1,2"],
            ["1,3"],
            ["2,3"],
            ["1,2,3"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(
            sample_distances,
            samples,
            PRECLUSTER_SIZE=3,
            MAX_COASSEMBLY_SAMPLES=3,
            )
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_size_three_of_four(self):
        sample_distances = pl.DataFrame([
            ["1", "2", 1-0.1],
            ["1", "3", 1-0.2],
            ["1", "4", 1-0.4],
            ["2", "3", 1-0.2],
            ["2", "4", 1-0.4],
            ["3", "4", 1-1],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["1", "2", "3", "4"])

        expected_clusters = pl.DataFrame([
            ["1,2"],
            ["1,3"],
            ["1,4"],
            ["2,3"],
            ["2,4"],
            # ["3,4"],
            ["1,2,3"],
            ["1,2,4"],
            # ["1,3,4"],
            # ["2,3,4"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(
            sample_distances,
            samples,
            PRECLUSTER_SIZE=3,
            MAX_COASSEMBLY_SAMPLES=3,
            )
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_size_three_of_four_missing(self):
        sample_distances = pl.DataFrame([
            ["1", "2", 1-0.1],
            ["1", "3", 1-0.2],
            ["2", "3", 1-0.2],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["1", "2", "3"])

        expected_clusters = pl.DataFrame([
            ["1,2"],
            ["1,3"],
            # ["1,4"],
            ["2,3"],
            # ["2,4"],
            # ["3,4"],
            ["1,2,3"],
            # ["1,2,4"],
            # ["1,3,4"],
            # ["2,3,4"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(
            sample_distances,
            samples,
            PRECLUSTER_SIZE=3,
            MAX_COASSEMBLY_SAMPLES=3,
            )
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_size_three_of_four_balanced(self):
        sample_distances = pl.DataFrame([
            ["1", "2", 1-0.2],
            ["1", "3", 1-0.3],
            ["1", "4", 1-0.4],
            ["2", "3", 1-1],
            ["2", "4", 1-0.2],
            ["3", "4", 1-0.1],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["1", "2", "3", "4"])

        expected_clusters = pl.DataFrame([
            ["1,2"],
            ["1,3"],
            # ["1,4"],
            # ["2,3"],
            ["2,4"],
            ["3,4"],
            ["1,2,3"],
            ["1,2,4"],
            ["1,3,4"],
            ["2,3,4"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(
            sample_distances,
            samples,
            PRECLUSTER_SIZE=3,
            MAX_COASSEMBLY_SAMPLES=3,
            )
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_size_three_of_four_balanced_missing(self):
        sample_distances = pl.DataFrame([
            ["1", "2", 1-0.2],
            ["1", "3", 1-0.3],
            ["1", "4", 1-0.4],
            ["2", "4", 1-0.2],
            ["3", "4", 1-0.1],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["1", "2", "3", "4"])

        expected_clusters = pl.DataFrame([
            ["1,2"],
            ["1,3"],
            # ["1,4"],
            # ["2,3"],
            ["2,4"],
            ["3,4"],
            ["1,2,3"],
            ["1,2,4"],
            ["1,3,4"],
            ["2,3,4"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(
            sample_distances,
            samples,
            PRECLUSTER_SIZE=3,
            MAX_COASSEMBLY_SAMPLES=3,
            )
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_size_four_of_five(self):
        sample_distances = pl.DataFrame([
            ["1", "2", 1-0.2],
            ["1", "3", 1-0.3],
            ["1", "4", 1-0.4],
            ["1", "5", 1-0.5],
            ["2", "3", 1-0.2],
            ["2", "4", 1-0.3],
            ["2", "5", 1-0.4],
            ["3", "4", 1-0.3],
            ["3", "5", 1-0.3],
            ["4", "5", 1-0.1],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["1", "2", "3", "4", "5"])

        expected_clusters = pl.DataFrame([
            ["1,2"],
            ["1,3"],
            ["1,4"],
            # ["1,5"],
            ["2,3"],
            ["2,4"],
            ["2,5"],
            ["3,4"],
            ["3,5"],
            ["4,5"],
            ["1,2,3"],
            ["1,2,4"],
            ["1,3,4"],
            ["2,3,4"],
            # ["1,2,5"],
            # ["1,3,5"],
            ["2,3,5"],
            # ["1,4,5"],
            ["2,4,5"],
            ["3,4,5"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(
            sample_distances,
            samples,
            PRECLUSTER_SIZE=4,
            MAX_COASSEMBLY_SAMPLES=3,
            )
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_size_four_of_five_missing(self):
        sample_distances = pl.DataFrame([
            ["1", "2", 1-0.2],
            ["1", "3", 1-0.3],
            ["1", "4", 1-0.4],
            ["2", "3", 1-0.2],
            ["2", "4", 1-0.3],
            ["2", "5", 1-0.4],
            ["3", "4", 1-0.3],
            ["3", "5", 1-0.3],
            ["4", "5", 1-0.1],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["1", "2", "3", "4", "5"])

        expected_clusters = pl.DataFrame([
            ["1,2"],
            ["1,3"],
            ["1,4"],
            # ["1,5"],
            ["2,3"],
            ["2,4"],
            ["2,5"],
            ["3,4"],
            ["3,5"],
            ["4,5"],
            ["1,2,3"],
            ["1,2,4"],
            ["1,3,4"],
            ["2,3,4"],
            # ["1,2,5"],
            # ["1,3,5"],
            ["2,3,5"],
            # ["1,4,5"],
            ["2,4,5"],
            ["3,4,5"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(
            sample_distances,
            samples,
            PRECLUSTER_SIZE=4,
            MAX_COASSEMBLY_SAMPLES=3,
            )
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_real_world(self):
        sample_distances = pl.DataFrame([
            ["SRR12149290", "SRR10571243", 1-0.16],
            ["ERR3201415", "ERR3220216", 1-0.16],
            ["ERR1414209", "ERR4804028", 1-0.16],
            ["SRR6979552", "SRR15213103", 1-0.07],
            ["SRR12149290", "SRR12352217", 1-0.16],
            ["SRR6979552", "SRR4831657", 1-0.09],
            ["ERR3201415", "SRR11784293", 1-0.09],
            ["SRR4249921", "SRR5207344", 1-0.15],
            ["SRR6979552", "SRR6980357", 1-0.25],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set([
            "SRR12149290", "SRR10571243",
            "ERR3201415", "ERR3220216",
            "ERR1414209", "ERR4804028",
            "SRR6979552", "SRR15213103",
            "SRR12149290", "SRR12352217",
            "SRR6979552", "SRR4831657",
            "ERR3201415", "SRR11784293",
            "SRR4249921", "SRR5207344",
            "SRR6979552", "SRR6980357",
        ])

        expected_clusters = pl.DataFrame([
            ["SRR10571243,SRR12149290"],
            ["ERR3201415,ERR3220216"],
            ["ERR1414209,ERR4804028"],
            ["SRR15213103,SRR6979552"],
            ["SRR12149290,SRR12352217"],
            ["SRR4831657,SRR6979552"],
            ["ERR3201415,SRR11784293"],
            ["SRR4249921,SRR5207344"],
            ["SRR6979552,SRR6980357"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(sample_distances, samples)
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_empty_inputs(self):
        sample_distances = pl.DataFrame([
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["sample_1", "sample_2"])

        expected_clusters = pl.DataFrame([
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(sample_distances, samples)
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_anchored(self):
        sample_distances = pl.DataFrame([
            ["sample_1", "sample_2", 1-0.5],
            ["sample_1", "sample_3", 1-1],
            ["sample_2", "sample_3", 1-0.9],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["sample_1", "sample_2", "sample_3"])
        anchor_samples = set(["sample_1"])

        expected_clusters = pl.DataFrame([
            ["sample_1,sample_2"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(sample_distances, samples, anchor_samples=anchor_samples)
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_anchored_size_four_of_five(self):
        sample_distances = pl.DataFrame([
            ["1", "2", 1-0.2],
            ["1", "3", 1-0.3],
            ["1", "4", 1-0.4],
            ["1", "5", 1-0.5],
            ["2", "3", 1-0.2],
            ["2", "4", 1-0.3],
            ["2", "5", 1-0.4],
            ["3", "4", 1-0.3],
            ["3", "5", 1-0.3],
            ["4", "5", 1-0.1],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set(["1", "2", "3", "4", "5"])
        anchor_samples = set(["1"])

        expected_clusters = pl.DataFrame([
            ["1,2"],
            ["1,3"],
            ["1,4"],
            # ["1,5"],
            # ["2,3"],
            # ["2,4"],
            # ["2,5"],
            # ["3,4"],
            # ["3,5"],
            # ["4,5"],
            ["1,2,3"],
            ["1,2,4"],
            ["1,3,4"],
            # ["2,3,4"],
            # ["1,2,5"],
            # ["1,3,5"],
            # ["2,3,5"],
            # ["1,4,5"],
            # ["2,4,5"],
            # ["3,4,5"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(
            sample_distances,
            samples,
            anchor_samples=anchor_samples,
            PRECLUSTER_SIZE=4,
            MAX_COASSEMBLY_SAMPLES=3,
            )
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_get_clusters_anchored_real_world(self):
        sample_distances = pl.DataFrame([
            ["SRR12149290", "SRR10571243", 1-0.16],
            ["ERR3201415", "ERR3220216", 1-0.16],
            ["ERR1414209", "ERR4804028", 1-0.16],
            ["SRR6979552", "SRR15213103", 1-0.07],
            ["SRR12149290", "SRR12352217", 1-0.16],
            ["SRR6979552", "SRR4831657", 1-0.09],
            ["ERR3201415", "SRR11784293", 1-0.09],
            ["SRR4249921", "SRR5207344", 1-0.15],
            ["SRR6979552", "SRR6980357", 1-0.25],
        ], orient="row", schema=SAMPLE_DISTANCES_COLUMNS)
        samples = set([
            "SRR12149290", "SRR10571243",
            "ERR3201415", "ERR3220216",
            "ERR1414209", "ERR4804028",
            "SRR6979552", "SRR15213103",
            "SRR12149290", "SRR12352217",
            "SRR6979552", "SRR4831657",
            "ERR3201415", "SRR11784293",
            "SRR4249921", "SRR5207344",
            "SRR6979552", "SRR6980357",
        ])
        anchor_samples = set(["SRR10571243", "SRR11784293", "SRR6979552"])

        expected_clusters = pl.DataFrame([
            ["SRR10571243,SRR12149290"],
            # ["ERR3201415,ERR3220216"],
            # ["ERR1414209,ERR4804028"],
            ["SRR15213103,SRR6979552"],
            # ["SRR12149290,SRR12352217"],
            # ["SRR4831657,SRR6979552"],
            ["ERR3201415,SRR11784293"],
            # ["SRR4249921,SRR5207344"],
            # ["SRR6979552,SRR6980357"],
        ], orient="row", schema=CLUSTERS_COLUMNS)

        observed_clusters = get_clusters(
            sample_distances,
            samples,
            anchor_samples=anchor_samples,
            )
        self.assertDataFrameEqual(expected_clusters, observed_clusters)

    def test_target_elusive(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", "0"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1,sample_2", "0"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_suffix(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_2.1", "AAA", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1.1", "sample_2.1"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_2.1", "AAA", 5, 10, "Root", "0"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1.1,sample_2.1", "0"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_suffix_underscore(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1_1", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_2_1", "AAA", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1_1", "sample_2_1"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1_1", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_2_1", "AAA", 5, 10, "Root", "0"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1_1,sample_2_1", "0"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_suffix_R(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1_R1", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_2_R1", "AAA", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1_R1", "sample_2_R1"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1_R1", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_2_R1", "AAA", 5, 10, "Root", "0"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1_R1,sample_2_R1", "0"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_empty_input(self):
        unbinned = pl.DataFrame([
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2"])

        expected_targets = pl.DataFrame([
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_no_matching_samples(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_2.1", "AAA", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_3", "sample_4"])

        expected_targets = pl.DataFrame([
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_low_coverage(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 5, "Root", ""],
            ["S3.1", "sample_2", "AAA", 5, 4, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 5, "Root", "0"],
            ["S3.1", "sample_2", "AAA", 5, 4, "Root", "0"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_two_targets(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_1", "AAB", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAB", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_1", "AAB", 5, 10, "Root", "1"],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_2", "AAB", 5, 10, "Root", "1"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1,sample_2", "0,1"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_multiple_genes(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", ""],
            ["S3.2", "sample_1", "AAB", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", ""],
            ["S3.2", "sample_2", "AAB", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", "0"],
            ["S3.2", "sample_1", "AAB", 5, 10, "Root", "1"],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", "0"],
            ["S3.2", "sample_2", "AAB", 5, 10, "Root", "1"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1,sample_2", "0,1"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_multiple_genes_same_sequence(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", ""],
            ["S3.2", "sample_1", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", ""],
            ["S3.2", "sample_2", "AAA", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", "0"],
            ["S3.2", "sample_1", "AAA", 5, 10, "Root", "1"],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", "0"],
            ["S3.2", "sample_2", "AAA", 5, 10, "Root", "1"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1,sample_2", "0,1"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_three_sample_mix(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_1", "AAB", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAB", 5, 10, "Root", ""],
            ["S3.1", "sample_3", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_3", "AAC", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2", "sample_3"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_1", "AAB", 5, 10, "Root", "1"],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_2", "AAB", 5, 10, "Root", "1"],
            ["S3.1", "sample_3", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_3", "AAC", 5, 10, "Root", "2"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1,sample_2", "0,1"],
            ["match", 2, "sample_1,sample_3", "0"],
            ["match", 2, "sample_2,sample_3", "0"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_skip_samples(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_1", "AAB", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAB", 5, 10, "Root", ""],
            ["S3.1", "sample_3", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_3", "AAC", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_1", "AAB", 5, 10, "Root", "1"],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_2", "AAB", 5, 10, "Root", "1"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1,sample_2", "0,1"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_three_way_targets(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 3, 6, "Root", ""],
            ["S3.1", "sample_1", "AAB", 2, 4, "Root", ""],
            ["S3.1", "sample_2", "AAA", 3, 6, "Root", ""],
            ["S3.1", "sample_2", "AAB", 2, 4, "Root", ""],
            ["S3.1", "sample_3", "AAA", 2, 4, "Root", ""],
            ["S3.1", "sample_3", "AAB", 2, 4, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2", "sample_3"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 3, 6, "Root", "0"],
            ["S3.1", "sample_1", "AAB", 2, 4, "Root", "1"],
            ["S3.1", "sample_2", "AAA", 3, 6, "Root", "0"],
            ["S3.1", "sample_2", "AAB", 2, 4, "Root", "1"],
            ["S3.1", "sample_3", "AAA", 2, 4, "Root", "0"],
            ["S3.1", "sample_3", "AAB", 2, 4, "Root", "1"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1,sample_2", "0"],
            ["pool", 3, "sample_1,sample_2,sample_3", "0,1"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples, MAX_COASSEMBLY_SAMPLES=3)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_four_way_targets(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 2, 4, "Root", ""],
            ["S3.1", "sample_1", "AAC", 1, 3, "Root", ""],
            ["S3.1", "sample_2", "AAA", 2, 4, "Root", ""],
            ["S3.1", "sample_2", "AAB", 2, 4, "Root", ""],
            ["S3.1", "sample_2", "AAC", 1, 3, "Root", ""],
            ["S3.1", "sample_3", "AAA", 2, 4, "Root", ""],
            ["S3.1", "sample_3", "AAB", 2, 4, "Root", ""],
            ["S3.1", "sample_3", "AAC", 1, 3, "Root", ""],
            ["S3.1", "sample_4", "AAB", 2, 4, "Root", ""],
            ["S3.1", "sample_4", "AAC", 1, 3, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2", "sample_3", "sample_4"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 2, 4, "Root", "0"],
            ["S3.1", "sample_1", "AAC", 1, 3, "Root", "1"],
            ["S3.1", "sample_2", "AAA", 2, 4, "Root", "0"],
            ["S3.1", "sample_2", "AAB", 2, 4, "Root", "2"],
            ["S3.1", "sample_2", "AAC", 1, 3, "Root", "1"],
            ["S3.1", "sample_3", "AAA", 2, 4, "Root", "0"],
            ["S3.1", "sample_3", "AAB", 2, 4, "Root", "2"],
            ["S3.1", "sample_3", "AAC", 1, 3, "Root", "1"],
            ["S3.1", "sample_4", "AAB", 2, 4, "Root", "2"],
            ["S3.1", "sample_4", "AAC", 1, 3, "Root", "1"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["pool", 3, "sample_1,sample_2,sample_3", "0"],
            ["pool", 3, "sample_2,sample_3,sample_4", "2"],
            ["pool", 4, "sample_1,sample_2,sample_3,sample_4", "1"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples, MAX_COASSEMBLY_SAMPLES=4)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    @unittest.skip("Benchmarking")
    def test_target_elusive_benchmarking(self):
        genes = ["S3.1", "S3.2", "S3.3", "S3.4", "S3.5", "S3.6", "S3.7", "S3.8", "S3.9", "S3.10"]
        samples = ["sample_" + str(n) for n in range(1, 51)]
        sequences = ["AAA", "AAB", "AAC", "AAD", "AAE", "AAF", "AAG", "AAH", "AAI", "AAJ"]
        unbinned = pl.DataFrame([
            [g, s, n, 2, 5.1, "Root", ""] for g in genes for s in samples for n in sequences
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(samples)

        _, _ = pipeline(unbinned, samples, MAX_COASSEMBLY_SAMPLES=4)
        _, _ = pipeline(unbinned, samples, MAX_COASSEMBLY_SAMPLES=4)
        _, _ = pipeline(unbinned, samples, MAX_COASSEMBLY_SAMPLES=4)
        _, _ = pipeline(unbinned, samples, MAX_COASSEMBLY_SAMPLES=4)
        _, _ = pipeline(unbinned, samples, MAX_COASSEMBLY_SAMPLES=4)

        # Cross join, recursive time: 247.833s
        # Pooling time: 1.979s

    def test_target_elusive_target_taxa(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root; d__Bacteria; p__Planctomycetota", ""],
            ["S3.1", "sample_1", "AAB", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root; d__Bacteria; p__Planctomycetota", ""],
            ["S3.1", "sample_2", "AAB", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root; d__Bacteria; p__Planctomycetota", "0"],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root; d__Bacteria; p__Planctomycetota", "0"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1,sample_2", "0"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples, TAXA_OF_INTEREST="p__Planctomycetota")
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_target_taxa_multiple(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root; d__Bacteria; p__Planctomycetota", ""],
            ["S3.1", "sample_1", "AAB", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root; d__Bacteria; p__Pseudomonadota", ""],
            ["S3.1", "sample_2", "AAB", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root; d__Bacteria; p__Planctomycetota", "0"],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root; d__Bacteria; p__Pseudomonadota", "0"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1,sample_2", "0"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples, TAXA_OF_INTEREST="p__Planctomycetota|p__Pseudomonadota")
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_single_assembly(self):
        unbinned = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_1", "AAB", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_2", "AAB", 5, 10, "Root", ""],
            ["S3.1", "sample_3", "AAA", 5, 10, "Root", ""],
            ["S3.1", "sample_3", "AAC", 5, 10, "Root", ""],
        ], orient="row", schema=APPRAISE_COLUMNS)
        samples = set(["sample_1", "sample_2", "sample_3"])

        expected_targets = pl.DataFrame([
            ["S3.1", "sample_1", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_1", "AAB", 5, 10, "Root", "1"],
            ["S3.1", "sample_2", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_2", "AAB", 5, 10, "Root", "1"],
            ["S3.1", "sample_3", "AAA", 5, 10, "Root", "0"],
            ["S3.1", "sample_3", "AAC", 5, 10, "Root", "2"],
        ], orient="row", schema=TARGETS_COLUMNS)
        expected_edges = pl.DataFrame([
            ["match", 2, "sample_1,sample_2", "0,1"],
            ["match", 2, "sample_1,sample_3", "0"],
            ["match", 2, "sample_2,sample_3", "0"],
        ], orient="row", schema=EDGES_COLUMNS)

        observed_targets, observed_edges = pipeline(unbinned, samples, MAX_COASSEMBLY_SAMPLES=1)
        self.assertDataFrameEqual(expected_targets, observed_targets)
        self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_preclustered_four_way(self):
        with in_tempdir():
            unbinned = pl.LazyFrame([
                ["S3.1", "sample_1", "AAA", 2, 4, "Root", ""],
                ["S3.1", "sample_1", "AAC", 1, 3.5, "Root", ""],
                ["S3.1", "sample_2", "AAA", 2, 4, "Root", ""],
                ["S3.1", "sample_2", "AAB", 2, 4, "Root", ""],
                ["S3.1", "sample_2", "AAC", 1, 3.5, "Root", ""],
                ["S3.1", "sample_3", "AAA", 2, 4, "Root", ""],
                ["S3.1", "sample_3", "AAB", 2, 4, "Root", ""],
                ["S3.1", "sample_3", "AAC", 1, 3.5, "Root", ""],
                ["S3.1", "sample_4", "AAB", 2, 4, "Root", ""],
                ["S3.1", "sample_4", "AAC", 1, 3.5, "Root", ""],
            ], orient="row", schema=APPRAISE_COLUMNS)
            samples = set(["sample_1", "sample_2", "sample_3", "sample_4"])
            preclusters = pl.DataFrame([
                ["sample_1,sample_3,sample_4"],
                ["sample_1,sample_2,sample_4"],
                ["sample_1,sample_2,sample_3,sample_4"],
            ], orient="row", schema=CLUSTERS_COLUMNS)

            expected_targets = pl.DataFrame([
                ["S3.1", "sample_1", "AAA", 2, 4, "Root", "3742656936210706615"],
                ["S3.1", "sample_1", "AAC", 1, 3.5, "Root", "4119220645959756749"],
                ["S3.1", "sample_2", "AAA", 2, 4, "Root", "3742656936210706615"],
                ["S3.1", "sample_2", "AAB", 2, 4, "Root", "3678881348843394362"],
                ["S3.1", "sample_2", "AAC", 1, 3.5, "Root", "4119220645959756749"],
                ["S3.1", "sample_3", "AAA", 2, 4, "Root", "3742656936210706615"],
                ["S3.1", "sample_3", "AAB", 2, 4, "Root", "3678881348843394362"],
                ["S3.1", "sample_3", "AAC", 1, 3.5, "Root", "4119220645959756749"],
                ["S3.1", "sample_4", "AAB", 2, 4, "Root", "3678881348843394362"],
                ["S3.1", "sample_4", "AAC", 1, 3.5, "Root", "4119220645959756749"],
            ], orient="row", schema=TARGETS_COLUMNS)
            expected_edges = pl.DataFrame([
                ["match", 3, "sample_1,sample_3,sample_4", "4119220645959756749"],
                ["match", 3, "sample_1,sample_2,sample_4", "4119220645959756749"],
                ["match", 4, "sample_1,sample_2,sample_3,sample_4", "4119220645959756749"],
            ], orient="row", schema=EDGES_COLUMNS)

            streaming_pipeline(
                unbinned,
                samples,
                sample_preclusters=preclusters,
                targets_path="targets.tsv",
                edges_path="edges.tsv",
                MAX_COASSEMBLY_SAMPLES=2,
                CHUNK_SIZE=2,
                )
            observed_targets = pl.read_csv("targets.tsv", schema_overrides=TARGETS_COLUMNS, separator="\t")
            observed_edges = pl.read_csv("edges.tsv", schema_overrides=EDGES_COLUMNS, separator="\t")
            self.assertDataFrameEqual(expected_targets, observed_targets)
            self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_preclustered_single_assembly(self):
        with in_tempdir():
            unbinned = pl.LazyFrame([
                ["S3.1", "sample_1", "AAA", 5, 10, "Root", ""],
                ["S3.1", "sample_1", "AAB", 5, 10, "Root", ""],
                ["S3.1", "sample_2", "AAA", 5, 10, "Root", ""],
                ["S3.1", "sample_2", "AAB", 5, 10, "Root", ""],
                ["S3.1", "sample_3", "AAA", 5, 10, "Root", ""],
                ["S3.1", "sample_3", "AAC", 5, 10, "Root", ""],
            ], orient="row", schema=APPRAISE_COLUMNS)
            samples = set(["sample_1", "sample_2", "sample_3"])
            preclusters = pl.DataFrame([
                # ["sample_1,sample_2"],
                ["sample_1,sample_3"],
                ["sample_2,sample_3"],
            ], orient="row", schema=CLUSTERS_COLUMNS)

            expected_targets = pl.DataFrame([
                ["S3.1", "sample_1", "AAA", 5, 10, "Root", "3742656936210706615"],
                ["S3.1", "sample_1", "AAB", 5, 10, "Root", "3678881348843394362"],
                ["S3.1", "sample_2", "AAA", 5, 10, "Root", "3742656936210706615"],
                ["S3.1", "sample_2", "AAB", 5, 10, "Root", "3678881348843394362"],
                ["S3.1", "sample_3", "AAA", 5, 10, "Root", "3742656936210706615"],
                ["S3.1", "sample_3", "AAC", 5, 10, "Root", "4119220645959756749"],
            ], orient="row", schema=TARGETS_COLUMNS)
            expected_edges = pl.DataFrame([
                # ["match", 2, "sample_1,sample_2", "3742656936210706615,3678881348843394362"],
                ["match", 2, "sample_1,sample_3", "3742656936210706615"],
                ["match", 2, "sample_2,sample_3", "3742656936210706615"],
            ], orient="row", schema=EDGES_COLUMNS)

            streaming_pipeline(
                unbinned,
                samples,
                sample_preclusters=preclusters,
                targets_path="targets.tsv",
                edges_path="edges.tsv",
                MAX_COASSEMBLY_SAMPLES=1,
                CHUNK_SIZE=2,
                )
            observed_targets = pl.read_csv("targets.tsv", schema_overrides=TARGETS_COLUMNS, separator="\t")
            observed_edges = pl.read_csv("edges.tsv", schema_overrides=EDGES_COLUMNS, separator="\t")
            self.assertDataFrameEqual(expected_targets, observed_targets)
            self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_preclustered_coalesce_target(self):
        with in_tempdir():
            unbinned = pl.LazyFrame([
                ["S3.1", "sample_1", "AAA", 5, 10, "Root", ""],
                ["S3.1", "sample_1", "AAB", 5, 10, "Root", ""],
                ["S3.1", "sample_2", "AAA", 5, 10, "Root", ""],
                ["S3.1", "sample_2", "AAB", 5, 10, "Root", ""],
                ["S3.1", "sample_3", "AAA", 5, 10, "Root", ""],
                ["S3.1", "sample_3", "AAC", 5, 10, "Root", ""],
            ], orient="row", schema=APPRAISE_COLUMNS)
            samples = set(["sample_1", "sample_2", "sample_3"])
            preclusters = pl.DataFrame([
                ["sample_1,sample_2"],
                ["sample_1,sample_3"],
                ["sample_2,sample_3"],
            ], orient="row", schema=CLUSTERS_COLUMNS)

            expected_targets = pl.DataFrame([
                ["S3.1", "sample_1", "AAA", 5, 10, "Root", "3742656936210706615"],
                ["S3.1", "sample_1", "AAB", 5, 10, "Root", "3678881348843394362"],
                ["S3.1", "sample_2", "AAA", 5, 10, "Root", "3742656936210706615"],
                ["S3.1", "sample_2", "AAB", 5, 10, "Root", "3678881348843394362"],
                ["S3.1", "sample_3", "AAA", 5, 10, "Root", "3742656936210706615"],
                ["S3.1", "sample_3", "AAC", 5, 10, "Root", "4119220645959756749"],
            ], orient="row", schema=TARGETS_COLUMNS)
            expected_edges = pl.DataFrame([
                ["match", 2, "sample_1,sample_2", "3678881348843394362,3742656936210706615"],
                ["match", 2, "sample_1,sample_3", "3742656936210706615"],
                ["match", 2, "sample_2,sample_3", "3742656936210706615"],
            ], orient="row", schema=EDGES_COLUMNS)

            streaming_pipeline(
                unbinned,
                samples,
                sample_preclusters=preclusters,
                targets_path="targets.tsv",
                edges_path="edges.tsv",
                MAX_COASSEMBLY_SAMPLES=1,
                CHUNK_SIZE=2,
                )
            observed_targets = pl.read_csv("targets.tsv", schema_overrides=TARGETS_COLUMNS, separator="\t")
            observed_edges = pl.read_csv("edges.tsv", schema_overrides=EDGES_COLUMNS, separator="\t")
            self.assertDataFrameEqual(expected_targets, observed_targets)
            self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_preclustered_no_targets(self):
        with in_tempdir():
            unbinned = pl.LazyFrame([
                ["S3.1", "sample_1", "AAA", 1, 3, "Root", ""],
                ["S3.2", "sample_1", "AAB", 1, 3, "Root", ""],
                ["S3.1", "sample_2", "AAA", 1, 3, "Root", ""],
                ["S3.2", "sample_2", "AAB", 1, 3, "Root", ""],
            ], orient="row", schema=APPRAISE_COLUMNS)
            samples = set(["sample_1", "sample_2"])
            preclusters = pl.DataFrame([
                ["sample_1,sample_2"],
            ], orient="row", schema=CLUSTERS_COLUMNS)

            expected_targets = pl.DataFrame([
                ["S3.1", "sample_1", "AAA", 1, 3, "Root", "3742656936210706615"],
                ["S3.2", "sample_1", "AAB", 1, 3, "Root", "376802398220639393"],
                ["S3.1", "sample_2", "AAA", 1, 3, "Root", "3742656936210706615"],
                ["S3.2", "sample_2", "AAB", 1, 3, "Root", "376802398220639393"],
            ], orient="row", schema=TARGETS_COLUMNS)
            expected_edges = pl.DataFrame([
            ], orient="row", schema=EDGES_COLUMNS)

            streaming_pipeline(
                unbinned,
                samples,
                sample_preclusters=preclusters,
                targets_path="targets.tsv",
                edges_path="edges.tsv",
                CHUNK_SIZE=2,
                )
            observed_targets = pl.read_csv("targets.tsv", schema_overrides=TARGETS_COLUMNS, separator="\t")
            observed_edges = pl.read_csv("edges.tsv", schema_overrides=EDGES_COLUMNS, separator="\t")
            self.assertDataFrameEqual(expected_targets, observed_targets)
            self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_preclustered_empty(self):
        with in_tempdir():
            unbinned = pl.LazyFrame([
                ["S3.1", "sample_1", "AAA", 2, 4, "Root", ""],
                ["S3.1", "sample_1", "AAC", 1, 3.5, "Root", ""],
                ["S3.1", "sample_2", "AAA", 2, 4, "Root", ""],
                ["S3.1", "sample_2", "AAB", 2, 4, "Root", ""],
                ["S3.1", "sample_2", "AAC", 1, 3.5, "Root", ""],
                ["S3.1", "sample_3", "AAA", 2, 4, "Root", ""],
                ["S3.1", "sample_3", "AAB", 2, 4, "Root", ""],
                ["S3.1", "sample_3", "AAC", 1, 3.5, "Root", ""],
                ["S3.1", "sample_4", "AAB", 2, 4, "Root", ""],
                ["S3.1", "sample_4", "AAC", 1, 3.5, "Root", ""],
            ], orient="row", schema=APPRAISE_COLUMNS)
            samples = set(["sample_1", "sample_2", "sample_3", "sample_4"])
            preclusters = pl.DataFrame([
            ], orient="row", schema=CLUSTERS_COLUMNS)

            expected_targets = pl.DataFrame([
                ["S3.1", "sample_1", "AAA", 2, 4, "Root", "3742656936210706615"],
                ["S3.1", "sample_1", "AAC", 1, 3.5, "Root", "4119220645959756749"],
                ["S3.1", "sample_2", "AAA", 2, 4, "Root", "3742656936210706615"],
                ["S3.1", "sample_2", "AAB", 2, 4, "Root", "3678881348843394362"],
                ["S3.1", "sample_2", "AAC", 1, 3.5, "Root", "4119220645959756749"],
                ["S3.1", "sample_3", "AAA", 2, 4, "Root", "3742656936210706615"],
                ["S3.1", "sample_3", "AAB", 2, 4, "Root", "3678881348843394362"],
                ["S3.1", "sample_3", "AAC", 1, 3.5, "Root", "4119220645959756749"],
                ["S3.1", "sample_4", "AAB", 2, 4, "Root", "3678881348843394362"],
                ["S3.1", "sample_4", "AAC", 1, 3.5, "Root", "4119220645959756749"],
            ], orient="row", schema=TARGETS_COLUMNS)
            expected_edges = pl.DataFrame([
            ], orient="row", schema=EDGES_COLUMNS)

            streaming_pipeline(
                unbinned,
                samples,
                sample_preclusters=preclusters,
                targets_path="targets.tsv",
                edges_path="edges.tsv",
                MAX_COASSEMBLY_SAMPLES=2,
                CHUNK_SIZE=2,
                )
            observed_targets = pl.read_csv("targets.tsv", schema_overrides=TARGETS_COLUMNS, separator="\t")
            observed_edges = pl.read_csv("edges.tsv", schema_overrides=EDGES_COLUMNS, separator="\t")
            self.assertDataFrameEqual(expected_targets, observed_targets)
            self.assertDataFrameEqual(expected_edges, observed_edges)

    def test_target_elusive_preclustered_empty_unbinned(self):
        with in_tempdir():
            unbinned = pl.LazyFrame([
            ], orient="row", schema=APPRAISE_COLUMNS)
            samples = set(["sample_1", "sample_2", "sample_3", "sample_4"])
            preclusters = pl.DataFrame([
            ], orient="row", schema=CLUSTERS_COLUMNS)

            expected_targets = pl.DataFrame([
            ], orient="row", schema=TARGETS_COLUMNS)
            expected_edges = pl.DataFrame([
            ], orient="row", schema=EDGES_COLUMNS)

            streaming_pipeline(
                unbinned,
                samples,
                sample_preclusters=preclusters,
                targets_path="targets.tsv",
                edges_path="edges.tsv",
                MAX_COASSEMBLY_SAMPLES=2,
                CHUNK_SIZE=2,
                )
            observed_targets = pl.read_csv("targets.tsv", schema_overrides=TARGETS_COLUMNS, separator="\t")
            observed_edges = pl.read_csv("edges.tsv", schema_overrides=EDGES_COLUMNS, separator="\t")
            self.assertDataFrameEqual(expected_targets, observed_targets)
            self.assertDataFrameEqual(expected_edges, observed_edges)


if __name__ == '__main__':
    unittest.main()
